#
# This file is part of LUNA
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" JTAG definitions for on-board JTAG devices. """

from .jtag import JTAGDevice


#
# ECP5 FPGAs that can be populated onto LUNA boards.
#

class LatticeECP5_12F(JTAGDevice):
    DEVICE = 'LFE5U-12F'
    DESCRIPTION = "Lattice LFE5U-12F ECP5 FPGA"
    SUPPORTED_IDCODES = [0x21111043]

class LatticeECP5_25F(JTAGDevice):
    DEVICE = 'LFE5U-25F'
    DESCRIPTION = "Lattice LFE5U-25F ECP5 FPGA"
    SUPPORTED_IDCODES = [0x41111043]

class LatticeECP5_45F(JTAGDevice):
    DEVICE = 'LFE5U-45F'
    DESCRIPTION = "Lattice LFE5U-45F ECP5 FPGA"
    SUPPORTED_IDCODES = [0x41112043]

class LatticeECP5_85F(JTAGDevice):
    DEVICE = 'LFE5U-85F'
    DESCRIPTION = "Lattice LFE5U-85F ECP5 FPGA"
    SUPPORTED_IDCODES = [0x41113043]


#
# Daisho boards.
#

class IntelCycloneIV_EP4CE30(JTAGDevice):
    """ Class representing a JTAG-connected CycloneIV, as on Daisho. """

    DEVICE      = "EP4CE30"
    DESCRIPTION = "Intel/Altera EP4CE30 Cyclone-IV FPGA"
    SUPPORTED_IDCODES = [0x020f40dd]



#
# Bit manipulation functions; lifted from Glasgow.
#   see: https://github.com/GlasgowEmbedded/glasgow/blob/master/software/glasgow/support/bits.py
#
# Copyright (C) 2018 whitequark@whitequark.org
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import re
import operator
from functools import reduce

import collections

try:
    import collections.abc
except ImportError:
    pass


__all__ = ["bits"]


class bits:
    """An immutable bit sequence, like ``bytes`` but for bits.

    This bit sequence is ordered from LSB to MSB; this is the direction in which it is converted
    to and from iterators, and to and from bytes. Note, however, that it is converted to and from
    strings (which should be only used where a human-readable form is required) from MSB to LSB;
    this matches the way integer literals are written, as well as values in datasheets and other
    documentation.
    """
    __slots__ = ["_len_", "_int_"]

    @classmethod
    def from_int(cls, value, length=None):
        value = operator.index(value)
        if length is None:
            if value < 0:
                raise ValueError("invalid negative input for bits(): '{}'".format(value))
            length = value.bit_length()
        else:
            length = operator.index(length)
            value &= ~(-1 << length)
        inst = object.__new__(cls)
        inst._len_ = length
        inst._int_ = value
        return inst

    @classmethod
    def from_str(cls, value):
        value  = re.sub(r"[\s_]", "", value)
        if value:
            if value[0] == "-":
                raise ValueError("invalid negative input for bits(): '{}'".format(value))
            elif value[0] == "+":
                length = len(value) - 1
            else:
                length = len(value)
            return cls.from_int(int(value, 2), length)
        else:
            return cls.from_int(0)

    @classmethod
    def from_iter(cls, iterator):
        length = -1
        value  = 0
        for length, bit in enumerate(iterator):
            value |= bool(bit) << length
        return cls.from_int(value, length + 1)

    @classmethod
    def from_bytes(cls, value, length, byteorder='little'):
        return cls.from_int(int.from_bytes(value, byteorder), length)

    def __new__(cls, value=0, length=None, byteorder='little'):
        if isinstance(value, cls):
            if length is None:
                return value
            else:
                return cls.from_int(value._int_, length)
        if isinstance(value, int):
            return cls.from_int(value, length)
        if isinstance(value, str):
            if length is not None:
                raise ValueError("invalid input for bits(): when converting from str "
                                 "length must not be provided")
            return cls.from_str(value)
        if isinstance(value, (bytes, bytearray, memoryview)):
            if length is None:
                raise ValueError("invalid input for bits(): when converting from bytes "
                                 "length must be provided")
            return cls.from_bytes(value, length, byteorder)
        if isinstance(value, collections.abc.Iterable):
            if length is not None:
                raise ValueError("invalid input for bits(): when converting from an iterable "
                                 "length must not be provided")
            return cls.from_iter(value)
        raise TypeError("invalid input for bits(): cannot convert from {}"
                        .format(value.__class__.__name__))

    def __len__(self):
        return self._len_

    def __bool__(self):
        return bool(self._len_)

    def to_int(self):
        return self._int_

    __int__ = to_int

    def to_str(self):
        if self._len_:
            return format(self._int_, "0{}b".format(self._len_))
        return ""

    __str__ = to_str

    def to_bytes(self, byteorder="little"):
        return self._int_.to_bytes((self._len_ + 7) // 8, byteorder)

    __bytes__ = to_bytes

    def __repr__(self):
        return "bits('{}')".format(self)

    def __getitem__(self, key):
        if isinstance(key, int):
            if key < 0:
                return (self._int_ >> (self._len_ - key)) & 1
            else:
                return (self._int_ >> key) & 1
        if isinstance(key, slice):
            start, stop, step = key.indices(self._len_)
            assert step == 1
            if stop < start:
                return self.__class__()
            else:
                return self.__class__(self._int_ >> start, stop - start)
        raise TypeError("bits indices must be integers or slices, not {}"
                        .format(key.__class__.__name__))

    def __iter__(self):
        for bit in range(self._len_):
            yield (self._int_ >> bit) & 1

    def __eq__(self, other):
        try:
            other = self.__class__(other)
        except TypeError:
            return False
        return self._len_ == other._len_ and self._int_ == other._int_

    def __add__(self, other):
        other = self.__class__(other)
        return self.__class__(self._int_ | (other._int_ << self._len_),
                              self._len_ + other._len_)

    def __radd__(self, other):
        other = self.__class__(other)
        return other + self

    def __mul__(self, other):
        if isinstance(other, int):
            return self.__class__(reduce(lambda a, b: (a << self._len_) | b,
                                         (self._int_ for _ in range(other)), 0),
                                  self._len_ * other)
        return NotImplemented

    def __rmul__(self, other):
        return self * other

    def __and__(self, other):
        other = self.__class__(other)
        return self.__class__(self._int_ & other._int_, max(self._len_, other._len_))

    def __rand__(self, other):
        other = self.__class__(other)
        return self & other

    def __or__(self, other):
        other = self.__class__(other)
        return self.__class__(self._int_ | other._int_, max(self._len_, other._len_))

    def __ror__(self, other):
        other = self.__class__(other)
        return self | other

    def __xor__(self, other):
        other = self.__class__(other)
        return self.__class__(self._int_ ^ other._int_, max(self._len_, other._len_))

    def __rxor__(self, other):
        other = self.__class__(other)
        return self ^ other

    def reversed(self):
        value = 0
        for bit in range(self._len_):
            value <<= 1
            if (self._int_ >> bit) & 1:
                value |= 1
        return self.__class__(value, self._len_)

# -------------------------------------------------------------------------------------------------

import unittest


class BitsTestCase(unittest.TestCase):
    def assertBits(self, value, bit_length, bit_value):
        self.assertIsInstance(value, bits)
        self.assertEqual(value._len_, bit_length)
        self.assertEqual(value._int_, bit_value)

    def test_from_int(self):
        self.assertBits(bits.from_int(0), 0, 0b0)
        self.assertBits(bits.from_int(1), 1, 0b1)
        self.assertBits(bits.from_int(2), 2, 0b10)
        self.assertBits(bits.from_int(2, 5), 5, 0b00010)
        self.assertBits(bits.from_int(0b110, 2), 2, 0b10)
        self.assertBits(bits.from_int(-1, 16), 16, 0xffff)

    def test_from_int_wrong(self):
        with self.assertRaisesRegex(ValueError,
                r"invalid negative input for bits\(\): '-1'"):
            bits.from_int(-1)

    def test_from_str(self):
        self.assertBits(bits.from_str(""), 0, 0b0)
        self.assertBits(bits.from_str("0"), 1, 0b0)
        self.assertBits(bits.from_str("010"), 3, 0b010)
        self.assertBits(bits.from_str("0 1  011_100"), 8, 0b01011100)
        self.assertBits(bits.from_str("+0 1 \t011_100"), 8, 0b01011100)

    def test_from_str_wrong(self):
        with self.assertRaisesRegex(ValueError,
                r"invalid negative input for bits\(\): '-1'"):
            bits.from_str("-1")
        with self.assertRaisesRegex(ValueError,
                r"invalid literal for int\(\) with base 2: '23'"):
            bits.from_str("23")

    def test_from_bytes(self):
        self.assertBits(bits.from_bytes(b"\xa5", 8), 8, 0b10100101)
        self.assertBits(bits.from_bytes(b"\xa5\x01", 9), 9, 0b110100101)
        self.assertBits(bits.from_bytes(b"\xa5\xff", 9), 9, 0b110100101)

    def test_from_iter(self):
        self.assertBits(bits.from_iter(iter([])), 0, 0b0)
        self.assertBits(bits.from_iter(iter([1,1,0,1,0,0,1])), 7, 0b1001011)

    def test_new(self):
        self.assertBits(bits(), 0, 0b0)
        self.assertBits(bits(10), 4, 0b1010)
        self.assertBits(bits(10, 2), 2, 0b10)
        self.assertBits(bits("1001"), 4, 0b1001)
        self.assertBits(bits(b"\xa5\x01", 9), 9, 0b110100101)
        self.assertBits(bits(bytearray(b"\xa5\x01"), 9), 9, 0b110100101)
        self.assertBits(bits(memoryview(b"\xa5\x01"), 9), 9, 0b110100101)
        self.assertBits(bits([1,1,0,1,0,0,1]), 7, 0b1001011)
        self.assertBits(bits(bits("1001"), 2), 2, 0b01)
        some = bits("1001")
        self.assertIs(bits(some), some)

    def test_new_wrong(self):
        with self.assertRaisesRegex(TypeError,
                r"invalid input for bits\(\): cannot convert from float"):
            bits(1.0)
        with self.assertRaisesRegex(ValueError,
                r"invalid input for bits\(\): when converting from str "
                r"length must not be provided"):
            bits("1010", 5)
        with self.assertRaisesRegex(ValueError,
                r"invalid input for bits\(\): when converting from bytes "
                r"length must be provided"):
            bits(b"\xa5")
        with self.assertRaisesRegex(ValueError,
                r"invalid input for bits\(\): when converting from an iterable "
                r"length must not be provided"):
            bits([1,0,1,0], 5)

    def test_len(self):
        self.assertEqual(len(bits(10)), 4)

    def test_bool(self):
        self.assertFalse(bits(""))
        self.assertTrue(bits("1"))
        self.assertTrue(bits("01"))
        self.assertTrue(bits("0"))
        self.assertTrue(bits("00"))

    def test_int(self):
        self.assertEqual(int(bits("1010")), 0b1010)

    def test_str(self):
        self.assertEqual(str(bits("")), "")
        self.assertEqual(str(bits("0000")), "0000")
        self.assertEqual(str(bits("1010")), "1010")
        self.assertEqual(str(bits("01010")), "01010")

    def test_bytes(self):
        self.assertEqual(bytes(bits("")), b"")
        self.assertEqual(bytes(bits("10100101")), b"\xa5")
        self.assertEqual(bytes(bits("110100101")), b"\xa5\x01")

    def test_repr(self):
        self.assertEqual(repr(bits("")), r"bits('')")
        self.assertEqual(repr(bits("1010")), r"bits('1010')")

    def test_getitem_int(self):
        some = bits("10001001011")
        self.assertEqual(some[0], 1)
        self.assertEqual(some[2], 0)
        self.assertEqual(some[5], 0)

    def test_getitem_slice(self):
        some = bits("10001001011")
        self.assertBits(some[:], 11, 0b10001001011)
        self.assertBits(some[2:], 9, 0b100010010)
        self.assertBits(some[2:9], 7, 0b0010010)
        self.assertBits(some[2:-2], 7, 0b0010010)
        self.assertBits(some[3:2], 0, 0b0)

    def test_getitem_wrong(self):
        with self.assertRaisesRegex(TypeError,
                r"bits indices must be integers or slices, not str"):
            bits()["x"]

    def test_iter(self):
        some = bits("10001001011")
        self.assertEqual(list(some), [1,1,0,1,0,0,1,0,0,0,1])

    def test_eq(self):
        self.assertEqual(bits("1010"), 0b1010)
        self.assertEqual(bits("1010"), "1010")
        self.assertEqual(bits("1010"), bits("1010"))
        self.assertNotEqual(bits("0010"), 0b0010)
        self.assertNotEqual(bits("0010"), "010")
        self.assertNotEqual(bits("1010"), bits("01010"))
        self.assertNotEqual(bits("1010"), None)

    def test_add(self):
        self.assertBits(bits("1010") + bits("1110"), 8, 0b11101010)
        self.assertBits(bits("1010") + (0,1,1,1), 8, 0b11101010)
        self.assertBits((0,1,1,1) + bits("1010"), 8, 0b10101110)

    def test_mul(self):
        self.assertBits(bits("1011") * 4, 16, 0b1011101110111011)
        self.assertBits(4 * bits("1011"), 16, 0b1011101110111011)

    def test_and(self):
        self.assertBits(bits("1010") & bits("1100"), 4, 0b1000)
        self.assertBits(bits("1010") & "1100", 4, 0b1000)
        self.assertBits((0,1,0,1) & bits("1100"), 4, 0b1000)

    def test_or(self):
        self.assertBits(bits("1010") | bits("1100"), 4, 0b1110)
        self.assertBits(bits("1010") | "1100", 4, 0b1110)
        self.assertBits((0,1,0,1) | bits("1100"), 4, 0b1110)

    def test_xor(self):
        self.assertBits(bits("1010") ^ bits("1100"), 4, 0b0110)
        self.assertBits(bits("1010") ^ "1100", 4, 0b0110)
        self.assertBits((0,1,0,1) ^ bits("1100"), 4, 0b0110)

    def test_reversed(self):
        self.assertBits(bits("1010").reversed(), 4, 0b0101)



#
# This file is part of LUNA
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from re import A
import sys
import logging

from warnings import warn

from .support.bits import bits
from .protocol.jtag_svf import SVFParser, SVFEventHandler

# Vendor requests that implement our basic JTAG protocol.
REQUEST_JTAG_START            = 0xbf
REQUEST_JTAG_CLEAR_OUT_BUFFER = 0xb0
REQUEST_JTAG_SET_OUT_BUFFER   = 0xb1
REQUEST_JTAG_GET_IN_BUFFER    = 0xb2
REQUEST_JTAG_SCAN             = 0xb3
REQUEST_JTAG_RUN_CLOCK        = 0xb4
REQUEST_JTAG_GO_TO_STATE      = 0xb5
REQUEST_JTAG_GET_STATE        = 0xb6
REQUEST_JTAG_GET_INFO         = 0xb8
REQUEST_JTAG_STOP             = 0xbe

#
# List of JTAG implementaiton quirks.
#

# Some serial engines can only send bytes from MSB to LSB, rather than
# the LSB to MSB that JTAG requires. Since we're typically using a serial
# engine to send whole bytes, this requires whole bytes to be flipped, but
# not any straggling bits. Setting this quirk will automatically handle this case.
QUIRK_FLIP_BITS_IN_WHOLE_BYTES = (1 << 0)

# In some cases, we may want to always bitbang our JTAG; e.g. if a device doesn't
# agree with the target's timing properties.
QUIRK_ALWAYS_BITBANG_JTAG      = (1 << 1)


class JTAGPatternError(IOError):
    """ Class for errors that come from a JTAG read not matching the expected response. """

    def __init__(self, message, result):
        self.result = result
        super(JTAGPatternError, self).__init__(message)


class JTAGDevice:
    """ Class representing a single device on a JTAG scan chain. """

    DESCRIPTION = "no description available"

    # A list of supported IDCODEs for the relevant class.
    # Used unless the supports_idcode() method is overridden.
    SUPPORTED_IDCODES = []

    @classmethod
    def from_idcode(cls, idcode, position_in_chain=0):
        """ Attempts to create a JTAGDevice object that fits the provided IDCODE. """

        # Assume the generic device class is the most appropriate class for the device, initially.
        most_appropriate_class = cls

        # Search each imported subclass for the
        for subclass in cls.__subclasses__():
            if subclass.supports_idcode(idcode):
                most_appropriate_class = subclass
                break


        # Finally, create an instance of the most appropriate class for this object.
        instance = object.__new__(most_appropriate_class)
        most_appropriate_class.__init__(instance, idcode, position_in_chain)

        return instance


    @classmethod
    def supports_idcode(cls, idcode):
        """
        Returns true iff this class supports the given IDCODE.

        This default implementation uses SUPPORTED_IDCODES, but subclasses can override this
        for more nuanced behavior.
        """
        return idcode in cls.SUPPORTED_IDCODES


    def idcode(self):
        """ Returns this device's IDCODE. """
        return self._idcode


    def description(self):
        """ Returns a short description of the device. """
        return self.DESCRIPTION


    def __init__(self, idcode, position_in_chain):
        self._idcode = idcode



class JTAGChain:
    """ Class representing a JTAG scan-chain interface. """

    # Short name for this type of interface.
    INTERFACE_SHORT_NAME = "jtag"

    STATE_NUMBERS = {
        'RESET':     0,
        'IDLE':      1,

        # Data register path.
        'DRSELECT':  2,
        'DRCAPTURE': 3,
        'DRSHIFT':   4,
        'DREXIT1':   5,
        'DRPAUSE':   6,
        'DREXIT2':   7,
        'DRUPDATE':  8,

        # Instruction register path.
        'IRSELECT':  9,
        'IRCAPTURE': 10,
        'IRSHIFT':   11,
        'IREXIT1':   12,
        'IRPAUSE':   13,
        'IREXIT2':   14,
        'IRUPDATE':  15
    }



    def __init__(self, debugger, max_frequency=405e3):
        """ Creates a new JTAG scan-chain interface.

        Paramters:
            board         -- the Apollo debugger we're working with.
            max_frequency -- the maximum frequency we should attempt scan out data with
        """

        # Store how many times the JTAG chain is being used; this allows us to open it
        # multiple times without re-init.
        self._reference_count = 0

        # Grab our JTAG API object.
        self.debugger = debugger

        # Configure our chain to run at the relevant frequency.
        self.frequency = int(max_frequency)

        # Default to 2048 bits per scan.
        # This will be overridden in many cases by the autodetection methods below.
        self.max_bits_per_scan = 256 * 8

        # By default, don't assume any quirks.
        self._bit_reverse_whole_bytes = False
        self._force_jtag_bitbang      = False


    def initialize(self):
        """
        Starts use of a persistent JTAG connection.
        When possible prefer using 'with' on this object.
        """
        self.__enter__()


    def __enter__(self):
        """ Starts a new JTAG communication. """

        self._reference_count += 1

        # If our JTAG chain was already in use, we don't need to initialize it.
        # Vacuously succeed.
        if self._reference_count > 1:
            return self

        # First, get our vitals on the JTAG connection...
        try:
            jtag_info = self.debugger.in_request(REQUEST_JTAG_GET_INFO, length=8)

            # .. including the maximum amount we're allowed to squish into a transaciton...
            self.max_bits_per_scan = int.from_bytes(jtag_info[0:4], byteorder='little') * 8

            # ... and any platform quirks present.
            quirks = int.from_bytes(jtag_info[4:], byteorder='little')
            self._bit_reverse_whole_bytes = bool(quirks & QUIRK_FLIP_BITS_IN_WHOLE_BYTES)

        except IOError:
            pass


        # Enable our JTAG comms.
        self.debugger.out_request(REQUEST_JTAG_START)

        # Move to the test/reset state.
        self.move_to_state('RESET')

        return self


    def __exit__(self, item_type, value, tb):
        """ Terminates an active JTAG communication. """

        # Free up our use of this hardware.
        if self._reference_count:
            self._reference_count -= 1

            # If the hardware is now no longer used, stop performing JTAG.
            if self._reference_count == 0:
                self.debugger.out_request(REQUEST_JTAG_STOP)


    def set_frequency(self, max_frequency):
        """ Sets the operating frequency of future transactions on this JTAG chain. """

        # FIXME: support this
        pass



    def initialize_chain(self):
        """ Put the scan chain into its initial state, allowing fresh JTAG communications. """

        # Pulse the TMS line five times -- this brings us into the TEST_RESET state, which resets the test logic.
        self._ensure_in_state('RESET')



    def _receive_data_chunk(self, bits_to_scan, advance_state=False):
        """ Performs a raw scan-in of data, and returns the result. """

        # Figure out how many whole bytes we'll need to read to get all of our bits.
        bytes_to_read = (bits_to_scan + 7) // 8

        # Perform the actual data scan-in...
        self.debugger.out_request(REQUEST_JTAG_SCAN, value=bits_to_scan, index=1 if advance_state else 0)

        # ... and then read the relevant data.
        result = self.debugger.in_request(REQUEST_JTAG_GET_IN_BUFFER, length=bytes_to_read)
        return bytes(result)



    def _receive_data(self, bits_to_scan):

        response = bytearray()
        self.debugger.out_request(REQUEST_JTAG_CLEAR_OUT_BUFFER)

        bits_left_to_scan = bits_to_scan
        while bits_left_to_scan > 0:
            bits_in_chunk = min(bits_left_to_scan, self.max_bits_per_scan)
            bits_left_to_scan -= bits_in_chunk
            chunk = self._receive_data_chunk(bits_in_chunk)

            response.extend(chunk)

        return self._reverse_bits_where_necessary(bits_to_scan, response)


    def _pad_data_to_length(self, length_in_bits, data=None):
        """ Pads a given data set to a given length, in bits. """

        # Compute how many bytes we need the data to be.
        target_length_bytes = (length_in_bits + 7) // 8

        # If our data doesn't need padding, return it directly.
        if data and (len(data) >= target_length_bytes):
            return data

        # Create a mutable array of data; and add any data we have.
        padded = bytearray()
        if data:
            padded.extend(data)

        # Figure out how much padding we need.
        padding_necessary = target_length_bytes - len(padded)
        padded.extend("b\0" * padding_necessary)

        # Return our padded data.
        return padded


    def _transmit_data(self, bits_to_scan, data=None):
        """ Performs a raw scan-out of data, discarding any result. """
        self._scan_data(bits_to_scan, data, ignore_response=True)


    def _scan_data(self, bits_to_scan, byte_data, ignore_response=False):
        """ Performs a raw scan-in of data, and returns the result. """

        total_scanned = 0
        total_to_scan = bits_to_scan

        receive  = bytearray()
        if byte_data:
            transmit = bytearray(byte_data)[::-1]

        #self._realign_data_if_necessary(bits_to_scan, transmit)

        bytes_per_chunk = self.max_bits_per_scan // 8
        bits_per_chunk  = bytes_per_chunk * 8

        while bits_to_scan:
            bits_in_chunk = min(bits_to_scan, bits_per_chunk)
            bits_to_scan -= bits_in_chunk
            advance_state = not bool(bits_to_scan)

            if byte_data:
                chunk = transmit[0:bytes_per_chunk]
                del transmit[0:bytes_per_chunk]
            else:
                chunk = None

            total_scanned += bits_in_chunk
            result = self._scan_data_chunk(bits_in_chunk, chunk, ignore_response, advance_state)

            receive.extend(result)

        return receive[::-1]



    def _reverse_bits_where_necessary(self, bits_to_scan, byte_data):
        """ Applies any neceessary bit-manipulations before/after a JTAG scan.

        Used as some hardware can only send MSB->LSB, and not LSB->MSB.
        """

        # Quick helper function to reverse the bits in our bitstream.
        def reverse_bits(num):
            binstr = "{:08b}".format(num)
            return int(binstr[::-1], 2)

        if not self._bit_reverse_whole_bytes:
            return byte_data

        # Reverse all bits in any whole bytes.
        bit_reversed = bytearray(byte_data)
        for i in range(bits_to_scan // 8):
            bit_reversed[i] = reverse_bits(bit_reversed[i])

        return bit_reversed


    def _scan_data_chunk(self, bits_to_scan, byte_data, ignore_response=False, advance_state=False):
        """ Performs a raw scan-in of data, and returns the result. """

        # Figure out how many whole bytes we'll need to read to get all of our bits.
        bytes_to_read = (bits_to_scan + 7) // 8

        # Perform our actual data scan-in.
        # TODO: break larger-than-maximum transactions into smaller ones.
        if byte_data:
            bytes_to_send = self._reverse_bits_where_necessary(bits_to_scan, byte_data)
            self.debugger.out_request(REQUEST_JTAG_SET_OUT_BUFFER, data=bytes_to_send)

        # Figure out the flags we're going to use for our transaction.
        flags  = 0b01 if advance_state            else 0
        flags |= 0b10 if self._force_jtag_bitbang else 0

        self.debugger.out_request(REQUEST_JTAG_SCAN, value=bits_to_scan, index=flags)

        if not ignore_response:
            result = self.debugger.in_request(REQUEST_JTAG_GET_IN_BUFFER, length=bytes_to_read)
            result = self._reverse_bits_where_necessary(bits_to_scan, result)
        else:
            result = b""

        return result



    def _ensure_in_state(self, state):
        """
        Ensures the JTAG TAP FSM is in the given state.
        If we're not; progresses the TAP FSM by pulsing TMS until we reach the relevant state.
        """

        state_number = self.STATE_NUMBERS[state]
        self.debugger.out_request(REQUEST_JTAG_GO_TO_STATE, value=state_number)

        # TODO: remove this; this if for debugging only
        current_state_raw = self.debugger.in_request(REQUEST_JTAG_GET_STATE, length=1)
        current_state_number = int.from_bytes(current_state_raw, byteorder='little')
        assert(current_state_number == state_number)


    def move_to_state(self, state_name):
        """ Moves the JTAG scan chain to the relevant state.

        Parameters:
            state_name: The target state to wind up in, as a string. States are accepted in the format
            defined in the JTAG SVF standard, and thus should be one of:
                "RESET", "IDLE", "DRSELECT", "DRCAPTURE", "DRSHIFT", "DREXIT1", "DRPAUSE",
                "DREXIT2", "DRUPDATE", "IRSELECT", "IRCAPTURE", "IRSHIFT", "IREXIT1", "IRPAUSE",
                "IREXIT2", "IRUPDATE"
        """
        self._ensure_in_state(state_name.strip())



    def _shift_while_in_state(self, state, tdi=None, length=None, ignore_response=False, byteorder='big'):
        """ Shifts data through the chain while in the given state. """

        # Normalize our data into a bitstring type that we can easily work with.
        # This both ensures we have a known format; and implicitly handles things like padding.
        if tdi:
            data_bits = bits(tdi, length, byteorder=byteorder)

            # Convert from our raw data to the format we'll need to send down to the device.
            bit_length = len(data_bits)
            data_bytes = data_bits.to_bytes(byteorder='big')

        else:
            if length is None:
                raise ValueError("either TDI or length must be provided!")

            bit_length = length

        # Move into our shift-DR state.
        self._ensure_in_state(state)

        # Finally, issue the transaction itself.
        if tdi and ignore_response:
            self._transmit_data(bit_length, data_bytes)
            return None
        elif tdi:
            result = self._scan_data(bit_length, data_bytes)
        else:
            result = self._receive_data(bit_length)

        # Return our data, converted back up to bits.
        return bits(result, bit_length, byteorder='big')


    def _validate_response(self, response_bits, tdo=None, mask=None):
        """ Validates the response provided by a _shift_while_in_state call, in the traditional JTAG SVF form. """

        # If we don't have any data to validate against, vacuously succeed.
        if (not tdo) or (not response_bits):
            return

        # If we have a mask, mask both the TDO value and response, and then compare.
        masked_response = mask & response_bits if mask else response_bits
        masked_tdo      = mask & tdo if mask else tdo

        if masked_response != masked_tdo:
            sys.stderr.flush()
            sys.stdout.flush()
            raise JTAGPatternError("Scan result did not match expected pattern: {} != {} (expected)!".format(
                    masked_response, masked_tdo), response_bits)


    def shift_data(self, tdi=None, length=None, tdo=None, mask=None,
            ignore_response=False, byteorder='big', state_after=None):
        """ Shifts data through the scan-chain's data register.

        Parameters:
            tdi    -- The bits to be scanned out via TDI. Can be a support.bits() object, a string of 1's and 0's,
                      an integer, or bytes. If this is an integer or bytes object, the length argument must be provided.
                      If omitted or None, a string of all zeroes will be used,
            length -- The length of the transaction to be performed, in bits. This can be longer than the TDI data;
                      in which case the transmission will be padded with zeroes.
            tdo    -- The expected data to be received from the scan operation. If this is provided, the read result
                      will be compared to this data (optionally masked by mask), and an exception will be thrown if
                      the data doesn't match this value. Designed to behave like the SVF TDO field.
            mask   -- If provided, the given tdo argument will be masked, such that only bits corresponding to a '1'
                      in this mask argument are considered when checking against 'tdo'. This is the behavior defiend
                      in the SVF standard; see it for more information.
            ignore_response -- If provided; the returned response will always be empty, and tdo and mask will be ignored.
                               This allows for slight a performance optimization, as we don't have to shuttle data back.
            byteorder       -- The byteorder to consider the tdi value in; if bytes are provided.

        Returns the bits read, or None if the response is ignored.
        """

        # Perform the core shift, and gather the response.
        response = self._shift_while_in_state('DRSHIFT', tdi=tdi, length=length, ignore_response=ignore_response,
                byteorder=byteorder)

        # Validate our response against any provided constraints.
        self._validate_response(response, tdo=tdo, mask=mask)

        if state_after:
            self.move_to_state(state_after)

        return response


    def shift_instruction(self, tdi=None, length=None, tdo=None, mask=None,
            ignore_response=False, byteorder='big', state_after=None):
        """ Shifts data through the chain's instruction register.

        Parameters:
            tdi    -- The bits to be scanned out via TDI. Can be a support.bits() object, a string of 1's and 0's,
                      an integer, or bytes. If this is an integer or bytes object, the length argument must be provided.
                      If omitted or None, a string of all zeroes will be used,
            length -- The length of the transaction to be performed, in bits. This can be longer than the TDI data;
                      in which case the transmission will be padded with zeroes.
            tdo    -- The expected data to be received from the scan operation. If this is provided, the read result
                      will be compared to this data (optionally masked by mask), and an exception will be thrown if
                      the data doesn't match this value. Designed to behave like the SVF TDO field.
            mask   -- If provided, the given tdo argument will be masked, such that only bits corresponding to a '1'
                      in this mask argument are considered when checking against 'tdo'. This is the behavior defiend
                      in the SVF standard; see it for more information.
            ignore_response -- If provided; the returned response will always be empty, and tdo and mask will be ignored.
                               This allows for slight a performance optimization, as we don't have to shuttle data back.
            byteorder       -- The byteorder to consider the tdi value in; if bytes are provided.

        Returns the bits read, or None if the response is ignored.
        """

        # Perform the core shift, and gather the response.
        response = self._shift_while_in_state('IRSHIFT', tdi=tdi, length=length, ignore_response=ignore_response,
                byteorder=byteorder)

        # Validate our response against any provided constraints.
        self._validate_response(response, tdo=tdo, mask=mask)

        if state_after:
            self.move_to_state(state_after)

        return response


    def run_test(self, cycles, from_state='IDLE', end_state=None):
        """ Places the device into the RUNTEST/IDLE (or provided) state, and pulses the JTAG clock.

        Paraameters:
            cycles -- The number of cycles for which the device should remain in the given state.
            from_state -- The state in which the cycles should be spent; defaults to IDLE.
            end_state -- The state in which the device should be placed after the test is complete.
        """

        if from_state:
            self.move_to_state(from_state)

        self.debugger.out_request(REQUEST_JTAG_RUN_CLOCK, cycles, 0)

        if end_state:
            self.move_to_state(end_state)


    def _create_device_for_idcode(self, idcode, position_in_chain):
        """ Creates a JTAGDevice object for the relevant idcode. """

        return JTAGDevice.from_idcode(idcode, position_in_chain)


    def enumerate(self, return_idcodes=False):
        """ Initializes the JTAG TAP FSM, and attempts to identify all connected devices.

        Parameters:
            return_idcodes -- If true, this method will return a list of IDCodes rather than JTAGDevice objects.

        Returns a list of JTAGDevices (return_idcodes=False) or JTAG IDCODES (return_idcodes=True).
        """

        devices = []

        # Place the JTAG TAP FSM into its initial state, so we can perform enumeration.
        self.initialize_chain()

        # Resetting the TAP FSM also automatically loaded the instruction register with the IDCODE
        # instruction, and accordingly filled the chain of data registers with each device's IDCODE.
        # We can accordingly just scan out the data using shift_data.

        # Once we (re-)initialize the chain, each device automatically loads the IDCODE instruction
        # for execution. This means that if we just scan in data, we'll receive each device's IDCODE,
        # followed by a null terminator (32 bits of zeroes).
        position_in_chain = 0
        while True:

            # Attempt to read a 32-bit IDCODE from the device.
            raw_idcode = self.shift_data(length=32)
            idcode = int.from_bytes(raw_idcode, byteorder='big')

            # If our IDCODE is all 1's, and we have no devices, we seem to be stuck at one.
            # Warn the user.
            if idcode == 0xFFFFFFFF and not devices:
                warn("TDO appears to be stuck at '1'. Check your wiring?")

            # If we've received our null IDCODE, we've finished enumerating the chain.
            # We'll also treat an all-1's IDCODE as a terminator, as this invalid IDCODE occurs
            # if TDI is stuck-at-one.
            if idcode in (0x00000000, 0xFFFFFFFF):
                break

            if return_idcodes:
                devices.append(idcode)
            else:
                devices.append(self._create_device_for_idcode(idcode, position_in_chain))

            position_in_chain += 1

        return devices


    def play_svf_instructions(self, svf_string, log_function=None, error_log_function=logging.error):
        """ Executes a string of JTAG SVF instructions, strumming the relevant scan chain.

        svf_string   -- A string containing valid JTAG SVF instructions to be executed.
        log_function -- If provided, this function will be called with verbose operation information.
        log_error    -- This function will be used to print information about errors that occur.
        """

        # Create the parser that will run our SVF file, and run our SVF.
        parser = SVFParser(svf_string, ApolloSVFEventHandler(self, log_function, error_log_function))
        parser.parse_file()


    def play_svf_file(self, svf_file, log_function=None, error_log_function=logging.error):
        """ Executes the JTAG SVF instructions from the given file.

        svf_file     -- A filename or file object pointing to a JTAG SVF file.
        log_function -- If provided, this function will be called with verbose operation information.
        log_error    -- This function will be used to print information about errors that occur.
        """

        close_after = False

        if isinstance(svf_file, str):
            svf_file = open(svf_file, 'r')
            close_after = True

        self.play_svf_instructions(svf_file.read(), log_function=log_function, error_log_function=error_log_function)

        if close_after:
            svf_file.close()


    def force_bitbang_mode(self, force_enable : bool):
        """ Controls whether we're forcing our device to communicate using bitbang mode. """
        self._force_jtag_bitbang = force_enable


class ApolloSVFEventHandler(SVFEventHandler):
    """ SVF event handler that delegates handling of SVF instructions to a GreatFET JTAG interface. """


    def __init__(self, interface, verbose_log_function=None, error_log_function=logging.error):
        """ Creates a new SVF event handler.

        Parameters:
            interface: The GreatFET JTAG interface that will execute our JTAG commands.
        """

        if verbose_log_function is None:
            verbose_log_function = lambda string : None
        if error_log_function is None:
            error_log_function = print

        self.interface = interface
        self.log = verbose_log_function
        self.log_error = error_log_function

        # Assume that after a data / instruction shift operation that we'll
        # wind up in the IDLE state, per the SVF standard. The SVF file can
        # override these defaults
        self.end_dr_state = 'IDLE'
        self.end_ir_state = 'IDLE'

        # By default, don't have any headers or trailers for IR or DR shifts.
        # The SVF can override these using the HDR/TDR/HIR/TIR instructions.
        nullary_padding = {'tdi': bits(), 'tdo': bits(), 'mask': bits(), }
        self.dr_header  = nullary_padding.copy()
        self.dr_trailer = nullary_padding.copy()
        self.ir_header  = nullary_padding.copy()
        self.ir_trailer = nullary_padding.copy()

        # Store default masks for our ShiftIR and ShiftDR instructions.
        self.last_dr_mask  = None
        self.last_dr_smask = None
        self.ir_mask  = None
        self.ir_smask = None


        self.interface.move_to_state('RESET')


    def svf_frequency(self, frequency):
        """Called when the ``FREQUENCY`` command is encountered."""
        self.log (" -- FREQUENCY set to {}".format(frequency))
        self.interface.set_frequency(frequency)


    def svf_trst(self, mode):
        """Called when the ``TRST`` command is encountered."""
        if mode != "ABSENT":
            warn('SVF provided TRST command; but this implementation does not yet support driving the TRST line')


    def svf_state(self, state, path):
        """Called when the ``STATE`` command is encountered."""

        # Visit each state in any intermediate paths provided...
        if path:
            for intermediate in path:
                self.log("STATE; Moving through {}.".format(intermediate))
                self.interface.move_to_state(intermediate)

        # ... ensuring we end up in the relevant state.
        self.log("Moving to {} STATE.".format(state))
        self.interface.move_to_state(state)


    def svf_endir(self, state):
        """Called when the ``ENDIR`` command is encountered."""
        self.log("Moving to {} after each Shift-IR.".format(state))
        self.end_dr_state = state


    def svf_enddr(self, state):
        """Called when the ``ENDDR`` command is encountered."""
        self.log("Moving to {} after each Shift-DR.".format(state))
        self.end_ir_state = state


    def svf_hir(self, **header):
        """Called when the ``HIR`` command is encountered."""
        self.log("Applying Shift-IR prefix. ")
        self.ir_header = header


    def svf_tir(self, **trailer):
        self.log("Applying Shift-IR suffix. ")
        self.ir_trailer = trailer


    def svf_hdr(self, **header):
        """Called when the ``HDR`` command is encountered."""
        self.log("Applying Shift-DR header. ")
        self.dr_header = header


    def svf_tdr(self, **trailer):
        """Called when the ``TDR`` command is encountered."""
        self.log("Applying Shift-DR suffix. ")
        self.dr_trailer = trailer


    def svf_sir(self, **data):
        """Called when the ``SIR`` command is encountered."""

        # Append our header and trailer to each of our arguments.
        arguments = {}
        for arg, value in data.items():
            header  = self.ir_header[arg] if (arg in self.ir_header) else bits()
            trailer = self.ir_trailer[arg] if (arg in self.ir_trailer) else bits()
            arguments[arg] = (header + value + trailer) if value else None

        if data['mask']:
            self.ir_mask = data['mask']
        if data['smask']:
            self.ir_smask = data['mask']

        self.log("Performing SHIFT-IR:")
        self.log(   "out:      {}".format(arguments['tdi']))
        self.log(   "expected: {}".format(arguments['tdo']))
        self.log(   "mask:     {}".format(arguments['mask']))
        try:
            result = self.interface.shift_instruction(tdi=arguments['tdi'], tdo=arguments['tdo'], mask=arguments['mask'])
        except JTAGPatternError as e:
            self.log(   "in:       {} [FAIL]\n".format(e.result))
            self.log_error("\n\n<!> Failure while performing SHIFT-IR: \n    " + str(e))
            raise

        self.log(   "in:       {} [OK]\n".format(result))


    def svf_sdr(self, **data):
        """Called when the ``SDR`` command is encountered."""

        # Append our header and trailer to each of our arguments.
        arguments = {}
        for arg, value in data.items():
            header  = self.dr_header[arg] if (arg in self.dr_header and self.dr_header[arg]) else bits()
            trailer = self.dr_trailer[arg] if (arg in self.dr_trailer and self.dr_header[arg]) else bits()
            arguments[arg] = (header + value + trailer) if value else None

        if data['mask']:
            self.dr_mask = data['mask']
        if data['smask']:
            self.dr_smask = data['mask']

        self.log("Performing SHIFT-DR:")
        self.log(   "out:      {}".format(arguments['tdi']))
        self.log(   "expected: {}".format(arguments['tdo']))
        self.log(   "mask:     {}".format(arguments['mask']))
        try:
            result = self.interface.shift_data(tdi=arguments['tdi'], tdo=arguments['tdo'], mask=arguments['mask'])
        except JTAGPatternError as e:
            self.log(   "in:       {} [FAIL]\n".format(e.result))
            self.log_error("\n\n<!> Failure while performing SHIFT-DR: \n    " + str(e))
            raise
        self.log(   "in:       {} [OK]\n".format(result))


    def svf_runtest(self, run_state, run_count, run_clock, min_time, max_time, end_state):
        """Called when the ``RUNTEST`` command is encountered."""
        self.log("Running test for {} cycles.".format(run_count))
        self.interface.run_test(run_count, from_state=run_state, end_state=end_state)


    def svf_piomap(self, mapping):
        """Called when the ``PIOMAP`` command is encountered."""
        raise NotImplementedError("This implementation does not yet support PIOMAP.")

    def svf_pio(self, vector):
        """Called when the ``PIO`` command is encountered."""
        raise NotImplementedError("This implementation does not yet support PIO.")



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Apollo-based ILA transports. """

from abc           import ABCMeta, abstractmethod

from vcd           import VCDWriter
from vcd.gtkw      import GTKWSave

from .support.bits import bits

class ILAFrontend(metaclass=ABCMeta):
    """ Class that communicates with an ILA module and emits useful output. """

    def __init__(self, ila):
        """
        Parameters:
            ila -- The ILA object to work with.
        """
        self.ila = ila
        self.samples = None


    @abstractmethod
    def _read_samples(self):
        """ Read samples from the target ILA. Should return an iterable of samples. """


    def _parse_sample(self, raw_sample):
        """ Converts a single binary sample to a dictionary of names -> sample values. """

        position = 0
        sample   = {}

        # Split our raw, bits(0) signal into smaller slices, and associate them with their names.
        for signal in self.ila.signals:
            signal_width = len(signal)
            signal_bits  = raw_sample[position : position + signal_width]
            position += signal_width

            sample[signal.name] = signal_bits

        return sample


    def _parse_samples(self, raw_samples):
        """ Converts raw, binary samples to dictionaries of name -> sample. """
        return [self._parse_sample(sample) for sample in raw_samples]


    def refresh(self):
        """ Fetches the latest set of samples from the target ILA. """
        self.samples = self._parse_samples(self._read_samples())


    def enumerate_samples(self):
        """ Returns an iterator that returns pairs of (timestamp, sample). """

        # If we don't have any samples, fetch samples from the ILA.
        if self.samples is None:
            self.refresh()

        timestamp = 0

        # Iterate over each sample...
        for sample in self.samples:
            yield timestamp, sample

            # ... and advance the timestamp by the relevant interval.
            timestamp += self.ila.sample_period


    def print_samples(self):
        """ Simple method that prints each of our samples; for simple CLI debugging."""

        for timestamp, sample in self.enumerate_samples():
            timestamp_scaled = 1000000 * timestamp
            print(f"{timestamp_scaled:08f}us: {sample}")



    def emit_vcd(self, filename, *, gtkw_filename=None, add_clock=True):
        """ Emits a VCD file containing the ILA samples.

        Parameters:
            filename      -- The filename to write to, or '-' to write to stdout.
            gtkw_filename -- If provided, a gtkwave save file will be generated that
                             automatically displays all of the relevant signals in the
                             order provided to the ILA.
            add_clock     -- If true or not provided, adds a replica of the ILA's sample
                             clock to make change points easier to see.
        """

        # Select the file-like object we're working with.
        if filename == "-":
            stream = sys.stdout
            close_after = False
        else:
            stream = open(filename, 'w')
            close_after = True

        # Create our basic VCD.
        with VCDWriter(stream, timescale=f"1 ns", date='today') as writer:
            first_timestamp = math.inf
            last_timestamp  = 0

            signals = {}

            # If we're adding a clock...
            if add_clock:
                clock_value  = 1
                clock_signal = writer.register_var('ila', 'ila_clock', 'integer', size=1, init=clock_value ^ 1)

            # Create named values for each of our signals.
            for signal in self.ila.signals:
                signals[signal.name] = writer.register_var('ila', signal.name, 'integer', size=len(signal))

            # Dump the each of our samples into the VCD.
            clock_time = 0
            for timestamp, sample in self.enumerate_samples():
                for signal_name, signal_value in sample.items():

                    # If we're adding a clock signal, add any changes necessary since
                    # the last value-change.
                    if add_clock:
                        while clock_time < timestamp:
                            writer.change(clock_signal, clock_time / 1e-9, clock_value)

                            clock_value ^= 1
                            clock_time  += (self.ila.sample_period / 2)

                    # Register the signal change.
                    writer.change(signals[signal_name], timestamp / 1e-9, signal_value.to_int())


        # If we're generating a GTKW, delegate that to our helper function.
        if gtkw_filename:
            assert(filename != '-')
            self._emit_gtkw(gtkw_filename, filename, add_clock=add_clock)


    def _emit_gtkw(self, filename, dump_filename, *, add_clock=True):
        """ Emits a GTKWave save file to accompany a generated VCD.

        Parameters:
            filename      -- The filename to write the GTKW save to.
            dump_filename -- The filename of the VCD that should be opened with this save.
            add_clock     -- True iff a clock signal should be added to the GTKW save.
        """

        with open(filename, 'w') as f:
            gtkw = GTKWSave(f)

            # Comments / context.
            gtkw.comment("Generated by the LUNA ILA.")

            # Add a reference to the dumpfile we're working with.
            gtkw.dumpfile(dump_filename)

            # If we're adding a clock, add it to the top of the view.
            gtkw.trace('ila.ila_clock')

            # Add each of our signals to the file.
            for signal in self.ila.signals:
                gtkw.trace(f"ila.{signal.name}")


    def interactive_display(self, *, add_clock=True):
        """ Attempts to spawn a GTKWave instance to display the ILA results interactively. """

        # Hack: generate files in a way that doesn't trip macOS's fancy guards.
        try:
            vcd_filename = os.path.join(tempfile.gettempdir(), os.urandom(24).hex() + '.vcd')
            gtkw_filename = os.path.join(tempfile.gettempdir(), os.urandom(24).hex() + '.gtkw')

            self.emit_vcd(vcd_filename, gtkw_filename=gtkw_filename)
            subprocess.run(["gtkwave", "-f", vcd_filename, "-a", gtkw_filename])
        finally:
            os.remove(vcd_filename)
            os.remove(gtkw_filename)


class AsyncSerialILAFrontend(ILAFrontend):
    """ UART-based ILA transport.

    Parameters
    ------------
    port: string
        The serial port to use to connect. This is typically a path on *nix systems.
    ila: IntegratedLogicAnalyzer
        The ILA object to work with.
    """

    def __init__(self, *args, ila, **kwargs):
        import serial

        self._port = serial.Serial(*args, **kwargs)
        self._port.reset_input_buffer()

        super().__init__(ila)


    def _split_samples(self, all_samples):
        """ Returns an iterator that iterates over each sample in the raw binary of samples. """

        sample_width_bytes = self.ila.bytes_per_sample

        # Iterate over each sample, and yield its value as a bits object.
        for i in range(0, len(all_samples), sample_width_bytes):
            raw_sample    = all_samples[i:i + sample_width_bytes]
            sample_length = len(Cat(self.ila.signals))

            yield bits.from_bytes(raw_sample, length=sample_length, byteorder='big')


    def _read_samples(self):
        """ Reads a set of ILA samples, and returns them. """

        sample_width_bytes = self.ila.bytes_per_sample
        total_to_read      = self.ila.sample_depth * sample_width_bytes

        # Fetch all of our samples from the given device.
        all_samples = self._port.read(total_to_read)
        return list(self._split_samples(all_samples))


if __name__ == "__main__":
    unittest.main()


class ApolloILAFrontend(ILAFrontend):
    """ Apollo-based transport for ILA samples. """

    def __init__(self, debugger, *, ila, use_inverted_cs=False):
        """
        Parameters:
            debugger        -- The apollo debugger connection to use for transport.
            ila             -- The ILA object to work with.
            use_inverted_cs -- Use a simple CS multiplexing scheme, where the ILA samples
                               are read out by pulsing SCK while CS is not asserted.
        """
        self._debugger = debugger
        self._use_inverted_cs = use_inverted_cs

        super().__init__(ila)


    def _split_samples(self, all_samples):
        """ Returns an iterator that iterates over each sample in the raw binary of samples. """

        from nmigen import Cat

        sample_width_bytes = self.ila.bytes_per_sample

        # Iterate over each sample, and yield its value as a bits object.
        for i in range(0, len(all_samples), sample_width_bytes):
            raw_sample    = all_samples[i:i + sample_width_bytes]
            sample_length = len(Cat(self.ila.signals))

            yield bits.from_bytes(raw_sample, length=sample_length, byteorder='big')


    def _read_samples(self):
        """ Reads a set of ILA samples, and returns them. """

        sample_width_bytes = self.ila.bytes_per_sample
        total_to_read      = self.ila.sample_depth * sample_width_bytes

        # Fetch all of our samples from the given device.
        all_samples = \
             self._debugger.spi.transfer(b"\0" * total_to_read, invert_cs=self._use_inverted_cs)

        return list(self._split_samples(all_samples))





#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

import usb.core

from .jtag  import JTAGChain
from .spi   import DebugSPIConnection
from .ila   import ApolloILAFrontend
from .ecp5  import ECP5_JTAGProgrammer, ECP5_JTAGDebugSPIConnection, ECP5_JTAGRegisters
from .intel import IntelJTAGProgrammer

from .onboard_jtag import *

class DebuggerNotFound(IOError):
    pass


def create_ila_frontend(ila, *, use_cs_multiplexing=False):
    """ Convenience method that instantiates an Apollo debug session and creates an ILA frontend from it.

    Parameters:
        ila -- The SyncSerialILA object we'll be connecting to.
    """
    debugger = ApolloDebugger()
    return ApolloILAFrontend(debugger, ila=ila, use_inverted_cs=use_cs_multiplexing)



class ApolloDebugger:
    """ Class representing a link to an Apollo Debug Module. """

    # This VID/PID pair is unique to development LUNA boards.
    # TODO: potentially change this to an OpenMoko VID, like other LUNA boards.
    USB_IDS  = [(0x1d50, 0x615c), (0x16d0, 0x05a5)]

    REQUEST_SET_LED_PATTERN = 0xa1
    REQUEST_RECONFIGURE     = 0xc0

    LED_PATTERN_IDLE = 500
    LED_PATTERN_UPLOAD = 50


    # External boards (non-LUNA boards) are indicated with a Major revision of 0xFF.
    # Their minor revision then encodes the board type.
    EXTERNAL_BOARD_MAJOR = 0xFF
    EXTERNAL_BOARD_NAMES = {
        0: "Daisho [rev 31-Oct-2014]",
        1: "Xil.se Pergola FPGA"
    }

    EXTERNAL_BOARD_PROGRAMMERS = {
        0: IntelJTAGProgrammer
    }


    # LUNA subdevices (specialized variants of the LUNA board) use a major of 0xFE.
    SUBDEVICE_MAJORS = {
        0xFE: "Amalthea"
    }


    def __init__(self):
        """ Sets up a connection to the debugger. """

        # Try to create a connection to our Apollo debug firmware.
        for vid, pid in self.USB_IDS:
            device = usb.core.find(idVendor=vid, idProduct=pid)
            if device is not None:
                break

        # If we couldn't find an Apollo device, bail out.
        if device is None:
            raise DebuggerNotFound()

        self.device = device
        self.major, self.minor = self.get_hardware_revision()

        # Create a basic JTAG chain interface, for debugging convenience.
        self.jtag  = JTAGChain(self)

        # Try to create an SPI-over-JTAG tunnel, if this board supports it.
        self.spi, self.registers = self.create_jtag_spi(self.jtag)

        # If it doesn't, use a hard SPI for JTAG-SPI.
        if self.spi is None:
            self.spi   = DebugSPIConnection(self)
            self.registers = self.spi



    def detect_connected_version(self):
        """ Attempts to determine the revision of the connected hardware.

        Returns the relevant hardware's revision number, as (major, minor).
        """

        # Extract the major and minor from the device's USB descriptor.
        minor = self.device.bcdDevice & 0xFF
        major = self.device.bcdDevice >> 8

        return major, minor


    def get_fpga_type(self):
        """ Returns a string indicating the type of FPGA populated on the connected LUNA board.

        The returned format is the same as used in a nMigen platform file; and can be used to override
        a platform's device type.
        """

        with self.jtag as jtag:

            # First, we'll detect all devices on our JTAG chain.
            jtag_devices = jtag.enumerate()
            if not jtag_devices:
                raise IOError("Could not detect an FPGA via JTAG!")

            # ... and grab its device identifier.
            first_device = jtag_devices[0]
            if not hasattr(first_device, 'DEVICE'):
                raise IOError("First JTAG device in chain does not provide an FPGA type. Is this a proper board?")

            return first_device.DEVICE


    @property
    def serial_number(self):
        """ Returns the device's serial number, as a string. """
        return self.device.serial_number


    def get_hardware_revision(self):
        """ Returns the (major, minor) of the attached hardware revision. """

        minor = self.device.bcdDevice & 0xFF
        major = self.device.bcdDevice >> 8
        return major, minor


    def get_hardware_name(self):
        """ Returns a string describing this piece of hardware. """

        # If this is a non-LUNA board, we'll look up its name in our table.
        if self.major == self.EXTERNAL_BOARD_MAJOR:
            return self.EXTERNAL_BOARD_NAMES[self.minor]

        # If this is a non-LUNA board, we'll look up its name in our table.
        if self.major in self.SUBDEVICE_MAJORS:
            product_name = self.SUBDEVICE_MAJORS[self.major]
            major        = 0 # For now?
        else:
            product_name = "LUNA"
            major        = self.major

        # Otherwise, identify it by its revision number.
        return f"{product_name} r{major}.{self.minor}"


    def get_compatibility_string(self):
        """ Returns 'LUNA' for a LUNA board; or 'LUNA-compatible' for supported external board."""

        if self.major == self.EXTERNAL_BOARD_MAJOR:
            return 'LUNA-compatible'
        elif self.major in self.SUBDEVICE_MAJORS:
            return self.SUBDEVICE_MAJORS[self.major]

        return 'LUNA'


    def create_jtag_programmer(self, jtag_chain):
        """ Returns the JTAG programmer for the given device. """

        # If this is an external programmer, return its programmer type.
        if self.major == self.EXTERNAL_BOARD_MAJOR:
            programmer = self.EXTERNAL_BOARD_PROGRAMMERS[self.minor]

        # Otherwise, it should be an ECP5.
        else:
            programmer = ECP5_JTAGProgrammer

        return programmer(jtag_chain)



    def create_jtag_spi(self, jtag_chain):
        """ Returns a JTAG-over-SPI connection for the given device. """

        # If this is an external programmer, we don't yet know how to create a JTAG-SPI 
        # interface for it. For now, assume we can't.
        if self.major == self.EXTERNAL_BOARD_MAJOR:
            return None, None

        # Otherwise, if we have a revision greater than r0.2, our SPI should be via JTAG.
        elif self.minor >= 0.3:
            return ECP5_JTAGDebugSPIConnection(jtag_chain, self), ECP5_JTAGRegisters(jtag_chain)

        # Otherwise, we'll want to use a real debug SPI, rather than a JTAG-SPI.
        else:
            return None, None



    def out_request(self, number, value=0, index=0, data=None, timeout=5000):
        """ Helper that issues an OUT control request to the debugger. """

        request_type = usb.ENDPOINT_OUT | usb.RECIP_DEVICE | usb.TYPE_VENDOR
        return self.device.ctrl_transfer(request_type, number, value, index, data, timeout=timeout)


    def in_request(self, number, value=0, index=0, length=0, timeout=500):
        """ Helper that issues an IN control request to the debugger. """

        request_type = usb.ENDPOINT_IN | usb.RECIP_DEVICE | usb.TYPE_VENDOR
        result = self.device.ctrl_transfer(request_type, number, value, index, length, timeout=timeout)

        return bytes(result)


    def set_led_pattern(self, number):
        self.out_request(self.REQUEST_SET_LED_PATTERN, number)


    def soft_reset(self):
        """ Resets the target (FPGA/etc) connected to the debug controller. """
        try:
            self.out_request(self.REQUEST_RECONFIGURE)
        except usb.core.USBError:
            pass



#
# This file is part of LUNA
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" ECP5 configuration code for LUNA. """

import time

from enum import IntEnum
from collections import defaultdict

from .jtag import JTAGChain
from .spi import DebugSPIConnection
from .support.bits import bits


class ECP5Programmer:
    """ Abstract base class for programming ECP5 FPGAs. """

    # Part names for various ECP5 part IDs.
    # Possibly missing a 12F variant of the 5UM. Definitely missing -5G part names.
    # TODO: grab all of the names from Trellis?
    PART_NAMES = {
        0x21111043: 'LFE5U-12',
        0x41111043: 'LFE5U-25',
        0x41112043: 'LFE5U-45',
        0x41113043: 'LFE5U-85',

        0x01111043: 'LFE5UM-25',
        0x01112043: 'LFE5UM-45',
        0x01113043: 'LFE5UM-25',
    }

    #
    # Status register components.
    #

    # JTAG mode is active.
    STATUS_FLAG_JTAG_ACTIVE = (1 << 4)

    # Configuration logic is busy.
    STATUS_FLAG_BUSY   = (1 << 12)

    # Last command failed.
    STATUS_FLAG_FAIL   = (1 << 13)

    # Done bit has been set.
    STATUS_FLAG_DONE   = (1 << 8)

    # Mismatch with VERIFY_ID command.
    STATUS_FLAG_ID_ERROR        = (1 << 27)

    # Invalid command detected, or execution failed.
    STATUS_FLAG_INVALID_COMMAND = (1 << 28)
    STATUS_FLAG_EXECUTION_FAIL = (1 << 26)

    # Programming status.
    STATUS_FLAG_ISC_ENABLE      = (1 <<  9)
    STATUS_FLAG_WRITEABLE       = (1 << 10)
    STATUS_FLAG_READABLE        = (1 << 11)
    STATUS_FLAG_STANDARD_PRE    = (1 << 21)

    # A previous SPI flash programming attempt failed.
    STATUS_FLAG_SPI_FAIL = (1 << 22)

    # Error codes.
    STATUS_ERROR_SHIFT = 23
    STATUS_ERROR_MASK  = 0b111
    STATUS_ERROR_CODES = {
        0: 'error unknown',
        1: 'part ID mismatch',
        2: 'illegal command issued',
        3: 'CRC check failed',
        4: 'preamble error',
        5: 'user aborted configuration',
        6: 'data overflow',
        7: 'bitstream provides data past the device\'s SRAM array',
    }


    # Length of an ECP5 part ID, in bytes.
    PART_ID_LENGTH = 4

    # Length of an ECP5 usercode, in bytes.
    USERCODE_LENGTH = 4

    # Length of an ECP5 status register, in bytes.
    STATUS_REGISTER_LENGTH = 4


    class Opcode(IntEnum):
        """ Enumeration that describes the ECP5 configuration opcodes. """

        NO_OP                                  = 0xFF

        # Read / verify the attached FPGA's part identifier.
        READ_ID                                = 0xE0
        VERIFY_ID                              = 0xE2

        # Read or program the device's usercode
        READ_USERCODE                          = 0xC0
        ISC_PROGRAM_USERCODE                   = 0xC2

        # Reads the attached FPGA's configuration status.
        LSC_READ_STATUS                        = 0x3C

        # Call REFRESH -- the equivalent of toggling the PROGRAMN pin.
        LSC_REFRESH                            = 0x79

        # Read whether the attached FPGA is still working on a command.
        LSC_CHECK_BUSY                         = 0xF0

        # Enables and disables configuration.
        ISC_ENABLE                             = 0xC6
        ISC_ENABLE_X                           = 0x74
        ISC_DISABLE                            = 0x26

        # Erase the device's SRAM.
        ISC_ERASE                              = 0x0E

        # Reset the internal CRC tracking, starting it fresh.
        LSC_RESET_CRC                          = 0x3B

        # Request that the device accept the data that follows as a contiguous bitstream.
        LSC_BITSTREAM_BURST                    = 0x7a

        # Configure the address to be programmed.
        LSC_SET_WORKING_ADDRESS                = 0x46

        # Configure the FPGA control registers.
        LSC_PROGRAM_CONTROL_REGISTER_0         = 0x22

        # Configures the FPGA with a block of data.
        LSC_PROGRAM_AND_INCREMENT_UNCOMPRESSED = 0x82
        LSC_PROGRAM_AND_INCREMENT_COMPRESSED   = 0xB8

        # Configure the FPGA's block RAM.
        LSC_SET_BLOCK_RAM_ADDRESS              = 0xF6
        LSC_SET_BLOCK_RAM_DATA                 = 0xB2

        # Configure how the device will work with an external SPI flash.
        SPI_MODE                               = 0x79

        # Mark FPGA configuration as complete.
        ISC_PROGRAM_DONE                       = 0x5E

        #
        # Opcodes known to Project Trellis, but unused.
        #
        JUMP                                   = 0b01111110
        LSC_WRITE_COMP_DIC                     = 0b00000010
        LSC_PROG_SED_CRC                       = 0b10100010
        ISC_PROGRAM_SECURITY                   = 0b11001110


    def __init__(self, cfg_pins=None, init_pin=None, program_pin=None, done_pin=None, verbose_function=None):
        """ Captures the common fields for all ECP5 programmers.

        Paramters, all optional:
            cfg_pins -- A (potentially virtual) port containing the three CFG pins that select the
                device's configuration mode. See VirtualGPIOPort for an example of a class that
                might be used for this.
            init_pin -- A GPIOPin object that connects the device's INIT_N (program status) pin.
            program_pin -- A GPIOPin object that drives the device's PROGRAM_B (program trigger) pin.
            done_pin -- A GPIOPin object that reads the status of the FPGA's done pin.

        If pins are not provided, their functionality (e.g. checking) will not be used, and it will
        be up to the subclass to handle this. Subclasses may choose to make these required.
        """

        # Store the pins for later use.
        self._cfg_pins    = cfg_pins
        self._program_pin = program_pin
        self._done_pin    = done_pin
        self._init_pin    = init_pin

        # Apply initial directions to each of our relevant pins.
        if self._cfg_pins:
            self._cfg_pins.all_output()
        if self._program_pin:
            self._program_pin.high()
        if self._done_pin:
            self._done_pin.input()
        if self._init_pin:
            self._init_pin.input()

        # Store our verbose print function.
        self._verbose_print = verbose_function if verbose_function else lambda m : None


    def _set_configuration_mode_pins(self, configuration_mode):
        """ Sets the configuration mode pins to the given binary-valued configuration mode.

        Returns True if this method was able to set the configration pins to their relevant value,
        or False otherwise.
        """

        if self._cfg_pins:
            self._cfg_pins.write(configuration_mode)
            return True
        else:
            return False



    def _restart_configuration_process(self):
        """ Attempts to strobe the PROGRAM line, to re-start the configuration process.

        Returns True iff we had the correct
        """

        if (not self._program_pin):
            return False

        # Strobe the PROGRAM_N pin, which triggers the device to accept new programming.
        self._program_pin.low()
        self._program_pin.high()

        # Wait for the device to be ready for new commands. We could poll INIT to see if
        # the initialization is complete, but TN1260 recommends we just wait 50ms.
        time.sleep(50 / 1000)



class ECP5CommandBasedProgrammer(ECP5Programmer):
    """
    Abstract base for ECP5 programmers that are command-centric (i.e. everything except Master SPI and Slave Serial.)
    To implement, override (at least) _execute_command.
    """


    def _execute_command(self, opcode, data_or_length=0, wait_for_completion=False, check_status=True, never_print=False):
        """ Issue an ECP5 configuration command.

        Parameters:
            opcode -- The opcode to issue.
            data_or_length -- A bytes-like object that contains the data to be sent; or an integer that contains
                the amount of data to be read.
            wait_for_completion -- If set, execution will block until the FPGA reports that the command is complete.
                Only necessary for long-running command.
            never_print -- True if this is a frequently-executed command (or a command called as part of this
                function's implementation). Prevents emission of verbose "what we're doing" text.
        """
        raise NotImplementedError("base classes of ECP5CommandBasedProgrammer must override _execute_command!")


    def read_id(self):
        """ Returns the identification code for the attached ECP5. """

        id_bytes = self._execute_command(self.Opcode.READ_ID, self.PART_ID_LENGTH, check_status=False)
        return int.from_bytes(id_bytes, byteorder='little')


    def part_name(self):
        """ Returns a descriptive name of the attached FPGA, if possible. """

        part_id = self.read_id()

        if part_id in self.PART_NAMES:
            return self.PART_NAMES[part_id]
        else:
            return "Unrecognized FPGA ({:08x})".format(part_id)



    def _validate_status(self, status, expect_done=False, continue_anyway=False, extra_verbose=False, expect_isc=False):
        """ Checks the status register after reading a given command. """

        def raise_error(msg):
            raise IOError(msg)

        def verbose_print(msg):
            if extra_verbose:
                self._verbose_print(msg)

        # Depending on whether we're continuing anyways, handle errors.
        if continue_anyway:
            error_handler = self._verbose_print
        else:
            error_handler = raise_error

        # Check for any error bits.
        error_code = (status >> self.STATUS_ERROR_SHIFT) & self.STATUS_ERROR_MASK
        error_text = " ({} / {:08x})".format(self.STATUS_ERROR_CODES[error_code], status)
        if status & self.STATUS_FLAG_FAIL:
            error_handler("Failed to execute last command!{}".format(error_text))
        if status & self.STATUS_FLAG_ID_ERROR:
            error_handler("Failed to verify device IDCODE!{}".format(error_text))
        if status & self.STATUS_FLAG_INVALID_COMMAND:
            error_handler("Last command was invalid!{}".format(error_text))
        #if error_code:
        #    error_handler("Error code reported. {}".format(error_text))

        # Print out some general status.
        if status:
            verbose_print("status was: {:08x}".format(status))
        if status & self.STATUS_FLAG_JTAG_ACTIVE:
            verbose_print("     - device is being controlled by JTAG.")
        if status & self.STATUS_FLAG_BUSY:
            verbose_print("     - device is busy.")
        if status & self.STATUS_FLAG_ISC_ENABLE:
            verbose_print("     - configuration enabled")
        if status & self.STATUS_FLAG_WRITEABLE:
            verbose_print("         - configuration can be written")
        if status & self.STATUS_FLAG_READABLE:
            verbose_print("         - configuration can be read back")
        if status & self.STATUS_FLAG_STANDARD_PRE:
            verbose_print("         - standard preamble detected")
        if status & self.STATUS_FLAG_SPI_FAIL:
            verbose_print("     - couldn't boot from SPI flash")
        if status & self.STATUS_FLAG_DONE:
            verbose_print("     - configuration is complete")

        # Check for DONE.
        if expect_done and not (status & self.STATUS_FLAG_DONE):
            error_handler("Configuration failed: {} / ({:08x})".format(self.STATUS_ERROR_CODES[error_code], status))


        if expect_isc and not (status & self.STATUS_FLAG_ISC_ENABLE):
            error_handler("Failed to enter ISC: {} / ({:08x})".format(self.STATUS_ERROR_CODES[error_code], status))


    def _perform_preconfiguration_tasks(self):
        """ Handles any tasks that need to be executed prior to configuration.

        Intended to allow subclasses a chance to add functionality. For example, a JTAG
        subclass would want to issue an LSC_PRELOAD here.
        """
        pass


    def _generate_bit_reversed_bitstream(self, bitstream, byte_reverse=False):
        """
        Generates a copy of the provided bitstream with the bits in each byte
        reversed -- in the format the FPGA likes them for MSPI mode.
        """

        # Quick helper function to reverse the bits in our bitstream.
        def reverse_bits(num):
            binstr = "{:08b}".format(num)
            return int(binstr[::-1], 2)

        # Reverse each of the bits in each byte of the bitstream.
        #
        # This ensures that bits are shifted into the FPGA in the same
        # order as they need to be presented to the configuration logic;
        # even if the FPGA is the one commanding the flash.
        #
        bit_reversed = bytearray(bitstream)
        for i in range(len(bit_reversed)):
            bit_reversed[i] = reverse_bits(bit_reversed[i])

        if byte_reverse:
            bit_reversed = bit_reversed[::-1]

        return bit_reversed


    def configure(self, bitstream):
        """ Configures the attached FPGA with the relevant bitstream file.

            bitstream -- An bytes-like object containing the data to be configured into the FPGA. Any object that
                can be passed to bytearray's constructor is acceptable.
        """

        bitstream = self._generate_bit_reversed_bitstream(bitstream, byte_reverse=True)

        self.chain.debugger.set_led_pattern(self.chain.debugger.LED_PATTERN_UPLOAD)

        try:
            # Ensure we're at the start of the configuration process. This also clears out any
            # existing bitstream.
            self._restart_configuration_process()

            # Perform any pre-configuration tasks necessary.
            self._perform_preconfiguration_tasks()

            # Capture the part ID, and then verify that our bitstream matches.
            # FIXME: use the bitstream file to get the ID, not our exected LUNA ID
            self._capture_part_id()
            #self._execute_command(self.Opcode.VERIFY_ID, b"\x21\x11\x10\x43")

            # ???
            self._execute_command(0x1C, bits(b"\x3f" + b"\xff" * 63, 510), check_status=False, bits_per_size_unit=1)

            # Enable configuration.
            self._execute_command(self.Opcode.ISC_ENABLE, b"\x00", check_status=False)
            self.chain.run_test(2)

            status = self._read_status()
            self._validate_status(status, expect_isc=True)

            # Erase the device's SRAM.
            self._execute_command(self.Opcode.ISC_ERASE, b"\x01", wait_for_completion=True, check_status=True)
            self.chain.run_test(2)

            # Check our status.
            status = self._read_status()
            self._validate_status(status, continue_anyway=True)

            # Shift the bitstream to the FPGA; this is essentially just executing all of the
            # commands in the bitstream.
            self._execute_command(self.Opcode.LSC_SET_WORKING_ADDRESS, b"\x01")

            # Shift the bitstream to the FPGA; this is essentially just executing all of the commands in the bitstream.
            self._execute_command(self.Opcode.LSC_BITSTREAM_BURST, bitstream, check_status=False, wait_for_completion=False)

            # Idle for long enough to let the configuration take.
            self._allow_configuration_time()

            # Check the device's status.
            status = self._read_status()
            self._validate_status(status, expect_done=True)

            # Disable configuration, and allow the FPGA to start.
            self._execute_command(self.Opcode.ISC_DISABLE, check_status=False)
            self.chain.run_test(2)

            status = self._read_status()
            self._validate_status(status, expect_done=True)


        finally:
            self.chain.debugger.set_led_pattern(self.chain.debugger.LED_PATTERN_IDLE)



    def _restart_configuration_process(self):
        """ Restarts the configuration process; equivalent to toggling the PROGRAM_N pin. """

        # First, try our base-most method of using the PROGRAM pin to restart configuration.
        parent_completed = super(ECP5CommandBasedProgrammer, self)._restart_configuration_process()

        # If this worked, we're done!
        if parent_completed:
            return

        # Otherwise, we'll attempt to do the same thing using the REFRESH command.

        # Note: for now, we're going to try to use the REFRESH command;
        # but it may make sense to reset the FPGA or use the PROGRAM_N pin.
        self._execute_command(self.Opcode.LSC_REFRESH, wait_for_completion=True, check_status=False)

        # Delay 50ms to give the board time to clear its SRAM.
        time.sleep(50 / 1000)


    def trigger_reconfiguration(self):
        pass


    def _capture_part_id(self):
        """ Reads and stores the device's part ID.

        Raises an IOError if the device has a known-invalid Device ID.
        """

        self.part_id = self.read_id()

        if self.part_id in (0, 0xFFFFFFFF):
            raise IOError("Could not detect a connected FPGA (ID: {:08x}). Check your wiring?".format(self.part_id))


    def _read_status(self) :
        """ Reads the part's status from the relevant device. """

        status_bytes = self._execute_command(self.Opcode.LSC_READ_STATUS, self.STATUS_REGISTER_LENGTH,
            check_status=False, never_print=True)
        return int.from_bytes(status_bytes, byteorder='big')


    def _read_usercode(self) :
        """ Reads the part's status from the relevant device. """

        status_bytes = self._execute_command(self.Opcode.READ_USERCODE, self.USERCODE_LENGTH)
        usercode =  int.from_bytes(status_bytes, byteorder='big')

        return usercode


    def _device_is_busy(self):
        """ Checks to see if the ECP5 device is busy handling a command. """

        busy = self._execute_command(self.Opcode.LSC_CHECK_BUSY, 1, check_status=False)
        return busy[0] & 0x01


    def _wait_for_completion(self, timeout=1, context=None):
        """ Blocks until the ECP5 reports that it has completed its current processing task.

        Parameters:
            timeout -- timeout in seconds
        """

        # Compute when we should time out.
        timeout = time.monotonic() + timeout

        # Wait until the device no longer reports that it's busy.
        while self._device_is_busy():
            time.sleep(0.01)

            if time.monotonic() > timeout:
                raise IOError("ECP5 request timed out waiting for completion (context: {})".format(context))



class ECP5SlaveSPI(ECP5CommandBasedProgrammer):
    """ Class that enables configuring ECP5 FPGAs via GreatFET boards. """


    # To enter Slave SPI mode, we'll need to set CFG[2:0] to 001.
    CONFIGURATION_SELECT_VALUE = 0b001


    def __init__(self, board, spi_bus=None, *args, **kwargs):
        """ Creates a new ECP5 Slave SPI (SSPI) configuration interface.

        Parameters:
            board -- The GreatFET board to use for configuration.
            spi_bus -- The SPI bus to use for programming. If omitted, the board's
                default SPI bus will be used.

        See ECP5Programmer.__init__ for additional accepted arguments.
        """

        # Store a reference to our board and SPI bus.
        self.board = board
        self.spi = spi_bus or board.spi

        # And run the parent configuration.
        super(ECP5SlaveSPI, self).__init__(*args, **kwargs)


    def _restart_configuration_process(self):
        """ Restarts the configuration process. """

        # Configure the FPGA's configuration select, and then call the parent function.
        self._set_configuration_mode_pins(self.CONFIGURATION_SELECT_VALUE)
        super(ECP5SlaveSPI, self)._restart_configuration_process()




    def _execute_command(self, opcode, data_or_length=0, wait_for_completion=False, check_status=True, never_print=False):
        """ Issue an ECP5 configuration command.

        Parameters:
            opcode -- The opcode to issue.
            data_or_length -- A bytes-like object that contains the data to be sent; or an integer that contains
                the amount of data to be read.
            wait_for_completion -- If set, execution will block until the FPGA reports that the command is complete.
                Only necessary for long-running command.
            never_print -- True if this is a frequently-executed command (or a command called as part of this
                function's implementation). Prevents emission of verbose "what we're doing" text.
        """

        # Start our command stream with our opcode, and our three required padding bytes.
        command_stream = bytearray([opcode, 0x00, 0x00, 0x00])
        prefix_length = len(command_stream)
        length_with_receive = len(command_stream)

        # ECP5 configuration transactions are either send -or- receive, but not both.
        # Accordingly, if we have a length, this is a receive-only transaction.
        is_receive = isinstance(data_or_length, int)

        # If we have a length, rather tha a data stream, compute the total transaction length.
        if is_receive:
            length_with_receive += data_or_length

        # Otherwise, we should have a data stream. Append it to our command stream, and then
        # issue that.
        else:
            length_with_receive += len(data_or_length)
            command_stream.extend(data_or_length)

        # Print out what we're doing, if we have a verbose function.
        if not never_print:
            operand_print = "" if len(command_stream) > 16 else ": {}".format(command_stream)
            self._verbose_print("Executing {} / [{} bytes{}]".format(self.Opcode(opcode).name,
                len(command_stream), operand_print))


        # Issue the command, and capture any data send in response.
        response = self.spi.transmit(command_stream, length_with_receive)

        # If the caller has requested we wait for completion, do so.
        if wait_for_completion:
            self._wait_for_completion(context=opcode)

        if check_status:
            status = self._read_status()
            self._validate_status(status)

        # If this is a receive command, trim off our command stream and return just the response.
        if is_receive:
            return bytes(response[prefix_length:])
        else:
            return b""


class ECP5MasterSerialDirect(ECP5Programmer):
    """ ECP5 Programmer class for programming the SPI flash used for Master SPI.

    This variant expects we have a direct connection to the target flash, rather than e.g.
    programming it through JTAG.
    """

    # To enter Master SPI mode, we'll need to set CFG[2:0] to 010.
    CONFIGURATION_SELECT_VALUE = 0b010


    def __init__(self, board, chip_select=None, *args, **kwargs):
        """ Creates a new ECP5 Slave SPI (SSPI) configuration interface.

        Parameters:
            board -- The GreatFET board to use for configuration.
            chip_select -- The GPIOPin to be used for chip select, or None to use the
                board's default.
            program_pin -- The GPIO pin to be used to control the FPGA's PROGRAM_N line.

        See ECP5Programmer.__init__ for additional accepted arguments.
        """

        # If we have a Chip Select string, convert it to a GPIOPin object.
        if isinstance(chip_select, str):
            chip_select = board.gpio.get_pin(chip_select)

        # If we do have a GPIOPin for chip select, make sure it's in input mode.
        if chip_select:
            chip_select.input()

        # Call our parent constructor and let it set things up.
        super(ECP5MasterSerialDirect, self).__init__(*args, **kwargs)

        # Create our connection to the target SPI flash.
        self.flash = SPIFlash(board, chip_select_port=chip_select.get_port(),
            chip_select_pin=chip_select.get_pin(), force_page_size=256)



    def trigger_reconfiguration(self):
        """ Triggers the target FPGA to reconfigure itself from its flash chip. """

        # Configure the FPGA's configuration select, and then call the parent function.
        self._set_configuration_mode_pins(self.CONFIGURATION_SELECT_VALUE)
        super(ECP5MasterSerialDirect, self)._restart_configuration_process()



    def program(self, bitstream, progress_callback=None):
        """ Programs the target flash with the relevant bitstream; but does not trigger reconfiguration of the FPGA. """
        bitstream_bit_reversed = self._generate_bit_reversed_bitstream(bitstream)
        self.flash.write(bitstream_bit_reversed, erase_first=True, progress_callback=progress_callback)


    def configure(self, bitstream, progress_callback=None):
        """ Programs the target flash, and then triggers the FPGA to reconfigure itself from it. """
        bitstream_bit_reversed = self._generate_bit_reversed_bitstream(bitstream)
        self.program(bitstream, progress_callback)
        self.trigger_reconfiguration()




class ECP5_JTAGProgrammer(ECP5CommandBasedProgrammer):
    """ Class that enables configuring ECP5 FPGAs over JTAG. """


    def __init__(self, jtag_chain, *args, **kwargs):
        """ Parameters:
            board -- The GreatFET board to use for configuration.
            chain -- The JTAG chain used for programming.
                default SPI bus will be used.

        See ECP5Programmer.__init__ for additional accepted arguments.
        """

        # Store a reference to our board and SPI bus.
        self.chain = jtag_chain

        # And run the parent configuration.
        super(ECP5_JTAGProgrammer, self).__init__(*args, **kwargs)


    def _allow_configuration_time(self):
        self.chain.shift_instruction(0xFF, state_after='IRPAUSE')
        self.chain.run_test(100)


    def _execute_command(self, opcode, data_or_length=0, wait_for_completion=False, check_status=True, never_print=False, bits_per_size_unit=8):
        """ Issue an ECP5 configuration command.

        Parameters:
           opcode -- The opcode to issue.
            data_or_length -- A bytes-like object that contains the data to be sent; or an integer that contains
                the amount of data to be read.
            wait_for_completion -- If set, execution will block until the FPGA reports that the command is complete.
                Only necessary for long-running command.
            never_print -- True if this is a frequently-executed command (or a command called as part of this
                function's implementation). Prevents emission of verbose "what we're doing" text.
        """

        # Scan in the opcode.
        self.chain.shift_instruction(opcode, state_after='IRPAUSE', length=8)

        # ECP5 configuration transactions are either send -or- receive, but not both.
        # Accordingly, if we have a length, this is a receive-only transaction.
        is_receive = isinstance(data_or_length, int)
        if is_receive:
            length = data_or_length * bits_per_size_unit
            data   = None
        else:
            data   = data_or_length
            length = len(data) * bits_per_size_unit


        # Issue the command, and capture any data send in response.
        if data or length:
            response = self.chain.shift_data(tdi=data, length=length, state_after='DRPAUSE')
        else:
            response = b""

        # If the caller has requested we wait for completion, do so.
        if wait_for_completion:
            self._wait_for_completion(context=opcode)

        if check_status:
            status = self._read_status()
            self._validate_status(status, extra_verbose=True)

        # If this is a receive command, return the response.
        if is_receive:
            return bytes(response)
        else:
            return b""






class ECP5_JTAGRegisters:
    """ JTAG-based connection for working with registers. """

    OPCODE_INSTRUCTION = 0x32
    OPCODE_DATA        = 0x38

    MAX_REGISTER_WIDTH = 128

    def __init__(self, jtag_chain):

        # Store our JTAG-specific parameters...
        self._chain = jtag_chain

        # Start off with no knowledge of our registers.
        self._instruction_width = None
        self._data_width        = None


    def _autodetect_width_for(self, opcode, chain):

        # Instruct the ECP5 to target the relevant register...
        chain.shift_instruction(opcode, state_after='IRPAUSE', length=8)

        # ... and shift out a long string of data.
        response = chain.shift_data(length=128, state_after='DRPAUSE').reversed()

        # Our registers start off filled with all 1's, in order to allow width detection.
        # We'll count how many 1's we see.
        for index, value in enumerate(response):
            if not value:
                return index

        # If we didn't see any -zeroes-, return 0, as this makes no sense to us.
        return 0

    def _autodetect_widths(self, chain):
        """ Attempts to automatically detect the instruction and data register widths. """

        # Ensure our JTAG chain is in a known state...
        self._chain.move_to_state('RESET')

        # And autodetect each of our widths.
        # Note that it's important to read our data width first, as scanning an instruction causes
        # the data register to be latched, losing our width.
        self._data_width = self._autodetect_width_for(self.OPCODE_DATA, chain)
        self._instruction_width = self._autodetect_width_for(self.OPCODE_INSTRUCTION, chain)

        # Sanity check our command-shape detection.
        invalid_shape = \
            (self._instruction_width      == 0) or \
            (self._data_width             == 0) or \
            (self._instruction_width % 8  != 0) or \
            (self._data_width % 8         != 0)

        if invalid_shape:
            raise IOError("Failed to autonegotiate meta-JTAG address/register size.")


    def _shift_with_opcode(self, chain, opcode, **arguments):

        # Instruct the ECP5 to target the register associated with our opcode...
        chain.shift_instruction(opcode, state_after='IRPAUSE', length=8)

        # ... and then shift out our data.
        result = chain.shift_data(**arguments, state_after='DRPAUSE')

        # Move to runtest for a bit to allow the target to process.
        chain.run_test(32)

        return result


    def _shift_instruction(self, chain, value):
        """ Shifts data into our instruction register via meta-JTAG. """
        return self._shift_with_opcode(chain, self.OPCODE_INSTRUCTION, tdi=value, length=self._instruction_width)


    def _shift_data(self, chain, value):
        """ Shifts data into our data register via meta-JTAG. """
        return self._shift_with_opcode(chain, self.OPCODE_DATA, tdi=value, length=self._data_width)


    def register_transaction(self, address, *, is_write, value=0):
        """ Performs an register transaction. """

        with self._chain as jtag:
            if (self._instruction_width is None) or (self._data_width is None):
                self._autodetect_widths(jtag)

            # Compute our write flag.
            write_flag_position = self._instruction_width - 1
            write_flag          = (1 << write_flag_position) if is_write else 0

            # Compute our command and value words.
            command = bits(value=write_flag | address, length=self._instruction_width, byteorder='big')
            value   = bits(value=value, length=self._data_width, byteorder='big')


            # Write our command into the instruction register...
            self._shift_instruction(jtag, command)

            # ... and shift our data.
            raw_result = self._shift_data(jtag, value)
            result = int(raw_result)

            return result


    def register_read(self, address):
        """ Reads a value from the provided registers."""
        return self.register_transaction(address, is_write=False)

    def register_write(self, address, value):
        """ Writes a value to the provided register. """
        return self.register_transaction(address, value=value, is_write=True)



class ECP5_JTAGDebugSPIConnection(DebugSPIConnection):
    """ JTAG-based soft-SPI connection for ECP5 boards. """

    OPCODE_ER1 = 0x32
    OPCODE_ER2 = 0x38

    def __init__(self, jtag_chain, debugger, opcode=OPCODE_ER1):

        # Store our JTAG-specific parameters...
        self._opcode = opcode
        self._chain = jtag_chain

        # ... and otherwise, initialize normally.
        super().__init__(debugger)


    def transfer(self, data_to_send, invert_cs=False):
        """ Transfers a set of data over SPI, and reads the response.

        Parameters:
            data_to_send -- The data to be sent; also sets the length of received data.
                            with a single CS line.
        """

        with self._chain as jtag:

            # Quick helper function to reverse bits in our response.
            def reverse_bits(num):
                binstr = "{:08b}".format(num)
                return int(binstr[::-1], 2)

            # Place the ECP5 into scan-through mode.
            jtag.shift_instruction(self._opcode, state_after='IRPAUSE', length=8)

            # Our JTAG protocol is bit-oriented; while our SPI is byte-oriented. In order to
            # abuse our JTAG transmission as SPI, we'll reverse the entire bit-stream (which places it
            # in an order such that the first byte is sent first); and then bit-reverse each byte (which
            # puts things in an MSB-first order, like SPI likes).
            byte_reversed_data = bytes(data_to_send)[::-1]
            jtag_ready_data    = bytes(reverse_bits(b) for b in byte_reversed_data)
            bits_to_send       = len(jtag_ready_data) * 8

            # Issue the command, and capture any data send in response.
            jtag.force_bitbang_mode(True)
            response = self._chain.shift_data(tdi=jtag_ready_data, length=bits_to_send, state_after='DRPAUSE')
            jtag.force_bitbang_mode(False)

            # Bit-reverse the data we capture in response, compensating for MSB-first ordering.
            response = [reverse_bits(b) for b in bytes(response)]
            return bytes(response)



#!/usr/bin/env python3
#
# This file is part of LUNA
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import print_function
from operator import invert

import os
import sys
import ast
import time
import errno
import logging
import argparse

from apollo_fpga import ApolloDebugger
from apollo_fpga.jtag import JTAGChain, JTAGPatternError
from apollo_fpga.ecp5 import ECP5_JTAGProgrammer
from apollo_fpga.onboard_jtag import *


COMMAND_HELP_TEXT = \
"""configure  -- Uploads a bitstream to the device's FPGA over JTAG.
program    -- Programs the target bitstream onto the attached FPGA.
jtag-scan  -- Prints information about devices on the onboard JTAG chain.
svf        -- Plays a given SVF file over JTAG.
spi        -- Sends the given list of bytes over debug-SPI, and returns the response.
spi-inv    -- Sends the given list of bytes over SPI with inverted CS.
spi-reg    -- Reads or writes to a provided register over the debug-SPI.
jtag-spi   -- Sends the given list of bytes over SPI-over-JTAG, and returns the response.
jtag-reg   -- Reads or writes to a provided register of JTAG-tunneled debug SPI.
"""


def print_device_info(device, args):
    """ Command that prints information about devices connected to the scan chain to the console. """

    logging.info(f"Detected a {device.get_compatibility_string()} device!")
    logging.info(f"\tHardware: {device.get_hardware_name()}")
    logging.info(f"\tSerial number: {device.serial_number}\n")


def print_chain_info(device, args):
    """ Command that prints information about devices connected to the scan chain to the console. """

    with device.jtag as jtag:
        logging.info("Scanning for connected devices...")
        detected_devices = jtag.enumerate()

        # If devices exist on the scan chain, print their information.
        if detected_devices:
            logging.info("{} device{} detected on the scan chain:\n".format(
                        len(detected_devices), 's' if len(detected_devices) > 1 else ''))

            for device in detected_devices:
                logging.info("    {:08x} -- {}".format(device.idcode(), device.description()))


            logging.info('')

        else:
            logging.info("No devices found.\n")


def play_svf_file(device, args):
    """ Command that prints the relevant flash chip's information to the console. """

    if not args.argument:
        logging.error("You must provide an SVF filename to play!\n")
        sys.exit(-1)

    with device.jtag as jtag:
        try:
            jtag.play_svf_file(args.argument)
        except JTAGPatternError:
            # Our SVF player has already logged the error to stderr.
            logging.error("")


def configure_fpga(device, args):
    """ Command that prints information about devices connected to the scan chain to the console. """


    with device.jtag as jtag:

        programmer = device.create_jtag_programmer(jtag)

        with open(args.argument, "rb") as f:
            bitstream = f.read()

        programmer.configure(bitstream)


def reconfigure_ecp5(device, args):
    """ Command that requests the attached ECP5 reconfigure itself from its MSPI flash. """

    device.soft_reset()



def _do_debug_spi(device, spi, args, *, invert_cs):

    # Try to figure out what data the user wants to send.
    data_raw = ast.literal_eval(args.argument)
    if isinstance(data_raw, int):
        data_raw = [data_raw]

    data_to_send = bytes(data_raw)
    response     = spi.transfer(data_to_send, invert_cs=invert_cs)

    print("response: {}".format(response))


def debug_spi(device, args, *, invert_cs=False):
    _do_debug_spi(device, device.spi, args, invert_cs=invert_cs)


def jtag_debug_spi(device, args):
    """ Command that issues data over a JTAG-over-SPI connection. """

    with device.jtag as jtag:
        spi, _ = device.create_jtag_spi(jtag)
        _do_debug_spi(device, spi, args, invert_cs=False)


def set_led_pattern(device, args):
    device.set_led_pattern(int(args.argument))

def debug_spi_inv(device, args):
    debug_spi(device, args, invert_cs=True)


def _do_debug_spi_register(device, spi, args):

    # Try to figure out what data the user wants to send.
    address = int(args.argument, 0)
    if args.value:
        value = int(args.value, 0)
        is_write = True
    else:
        value = 0
        is_write = False

    try:
        response = spi.register_transaction(address, is_write=is_write, value=value)
        print("0x{:08x}".format(response))
    except IOError as e:
        logging.critical(f"{e}\n")


def debug_spi_register(device, args):
    _do_debug_spi_register(device, device.spi, args)

def jtag_debug_spi_register(device, args):
    _reg, reg = device.create_jtag_spi(device.jtag)
    _do_debug_spi_register(device, reg, args)


def main():

    commands = {
        # Info queries
        'info':        print_device_info,
        'jtag-scan':   print_chain_info,

        # JTAG commands
        'svf':         play_svf_file,
        'configure':   configure_fpga,
        'reconfigure': reconfigure_ecp5,

        # SPI debug exchanges
        'spi':         debug_spi,
        'spi-inv':     debug_spi_inv,
        'spi-reg':     debug_spi_register,

        # JTAG-SPI debug exchanges.
        'jtag-spi':    jtag_debug_spi,
        'jtag-reg':    jtag_debug_spi_register,

        # Misc
        'leds':        set_led_pattern,

    }


    # Set up a simple argument parser.
    parser = argparse.ArgumentParser(description="Apollo FPGA Configuration / Debug tool",
            formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('command', metavar='command:', choices=commands, help=COMMAND_HELP_TEXT)
    parser.add_argument('argument', metavar="[argument]", nargs='?',
                        help='the argument to the given command; often a filename')
    parser.add_argument('value', metavar="[value]", nargs='?',
                        help='the value to a register write command')

    args = parser.parse_args()
    device = ApolloDebugger()

    # Set up python's logging to act as a simple print, for now.
    logging.basicConfig(level=logging.INFO, format="%(message)-s")

    # Execute the relevant command.
    command = commands[args.command]
    command(device, args)


if __name__ == '__main__':
    main()



#
# SVF protocol driver; lifted from Glasgow with minor modifications.
#   see: https://github.com/GlasgowEmbedded/glasgow/blob/master/software/glasgow/protocol/jtag_svf.py
#
# Copyright (C) 2018 whitequark@whitequark.org
#
# Permission to use, copy, modify, and/or distribute this software for
# any purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
# AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


import re
from abc import ABCMeta, abstractmethod

from ..support.bits import bits


__all__ = ["SVFParser", "SVFEventHandler"]


def _hex_to_bits(input_nibbles):
    return bits(int(input_nibbles, 16))


_commands = (
    "ENDDR", "ENDIR", "FREQUENCY", "HDR", "HIR", "PIO", "PIOMAP", "RUNTEST",
    "SDR", "SIR", "STATE", "TDR", "TIR", "TRST",
)
_parameters = (
    "ENDSTATE", "HZ", "MASK", "MAXIMUM", "SCK", "SEC", "SMASK", "TCK", "TDI", "TDO",
)
_trst_modes = (
    "ON", "OFF", "Z", "ABSENT"
)
_tap_states = (
    "RESET", "IDLE", "DRSELECT", "DRCAPTURE", "DRSHIFT", "DREXIT1", "DRPAUSE",
    "DREXIT2", "DRUPDATE", "IRSELECT", "IRCAPTURE", "IRSHIFT", "IREXIT1", "IRPAUSE",
    "IREXIT2", "IRUPDATE",
)
_tap_stable_states = (
    "RESET", "IDLE", "IRPAUSE", "DRPAUSE"
)


class SVFParsingError(Exception):
    pass


class SVFLexer:
    """
    A Serial Vector Format lexer.

    Comments (``! comment``, ``// comment``) are ignored.

    The following tokens are recognized:
        * Keyword (``HIR``, ``SIR``, ``TIO``, ..., ``;``), returned as Python ``str``;
        * Integer (``8``, ``16``, ...), returned as Python ``int``;
        * Real (``1E0``, ``1E+0``, ``1E-0``, ...), returned as Python ``float``;
        * Bit array (``(0)``, ``(1234)``, ``(F00F)``, ...), returned as Python ``bits``;
        * Literal (``(HLUDXZHHLL)``, ``(IN FOO)``, ...), returned as Python ``tuple(str,)``;
        * End of file, returned as Python ``None``.

    :type buffer: str
    :attr buffer:
        Input buffer.

    :type position: int
    :attr position:
        Offset into buffer from which the next token will be read.
    """

    _keywords = _commands + _parameters + _trst_modes + _tap_states + (";",)
    _scanner  = tuple((re.compile(src, re.UNICODE|re.I|re.M), act) for src, act in (
        (r"\s+",
         None),
        (r"(?:!|//)([^\n]*)(?:\n|\Z)",
         None),
        (r"({})(?=\s+|[;()]|\Z)".format("|".join(_keywords)),
         lambda m: m[1]),
        (r"(\d+)(?![\.E])",
         lambda m: int(m[1])),
        (r"(\d+(?:\.\d+)?(?:E[+-]?\d+)?)",
         lambda m: float(m[1])),
        (r"\(\s*([0-9A-F\s]+)\s*\)",
         lambda m: _hex_to_bits(re.sub(r"\s+", "", m[1]))),
        (r"\(\s*(.+?)\s*\)",
         lambda m: (m[1],)),
        (r"\Z",
         lambda m: None),
    ))

    def __init__(self, buffer):
        self.buffer   = buffer
        self.position = 0

    def line_column(self, position=None):
        """
        Return a ``(line, column)`` tuple for the given or, if not specified, current position.

        Both the line and the column start at 1.
        """
        line = len(re.compile(r"\n").findall(self.buffer, endpos=self.position))
        if line > 1:
            column = self.position - self.buffer.rindex("\n", 0, self.position)
        else:
            column = self.position
        return line + 1, column + 1

    def _lex(self):
        while True:
            for token_re, action in self._scanner:
                match = token_re.match(self.buffer, self.position)
                # print(token_re, match)
                if match:
                    if action is None:
                        self.position = match.end()
                        break
                    else:
                        return action(match), match.end()
            else:
                line, column = self.line_column()
                print_args = (line, column, self.buffer[self.position:self.position + 16])
                raise SVFParsingError("unrecognized SVF data at line {}, column {} ({}...)".format(print_args))

    def peek(self):
        """Return the next token without advancing the position."""
        token, _ = self._lex()
        return token

    def next(self):
        """Return the next token and advance the position."""
        token, next_pos = self._lex()
        self.position = next_pos
        return token

    def __iter__(self):
        return self

    def __next__(self):
        token = self.next()
        if token is None:
            raise StopIteration
        return token


class SVFParser:
    """
    A Serial Vector Format streaming parser.

    This parser maintains and allows querying lexical state (e.g. "sticky" ``TDI`` is
    automatically tracked), and invokes the SVF event handler for all commands so that
    any necessary action may be taken.
    """
    def __init__(self, buffer, handler):
        self._lexer     = SVFLexer(buffer)
        self._handler   = handler
        self._position  = 0
        self._token     = None
        self._cmd_pos   = 0

        self._param_tdi   = \
            {"HIR": None, "HDR": None, "SIR": None, "SDR": None, "TIR": None, "TDR": None}
        self._param_mask  = \
            {"HIR": None, "HDR": None, "SIR": None, "SDR": None, "TIR": None, "TDR": None}
        self._param_smask = \
            {"HIR": None, "HDR": None, "SIR": None, "SDR": None, "TIR": None, "TDR": None}

        self._param_run_state = "IDLE"
        self._param_end_state = "IDLE"

    def _try(self, action, *args):
        try:
            old_position = self._lexer.position
            return action(*args)
        except SVFParsingError:
            self._lexer.position = old_position
            return None

    def _parse_token(self):
        self._position = self._lexer.position
        self._token    = self._lexer.next()
        # print("token %s @ %d" % (self._token, self._position))
        return self._token

    def _parse_error(self, error):
        line, column = self._lexer.line_column(self._position)
        raise SVFParsingError("%s at line %d, column %d" % (error, line, column))

    def _parse_unexpected(self, expected, valid=()):
        if isinstance(self._token, str):
            actual = self._token
        elif isinstance(self._token, int):
            actual = "integer"
        elif isinstance(self._token, float):
            actual = "real"
        elif isinstance(self._token, bits):
            actual = "scan data"
        elif isinstance(self._token, tuple):
            actual = "(%s)" % tuple(self._token)
        elif self._token is None:
            actual = "end of file"
        else:
            assert False
        if valid:
            self._parse_error("expected %s (one of %s), found %s"
                              % (expected, ", ".join(valid), actual))
        else:
            self._parse_error("expected %s, found %s"
                              % (expected, actual))

    def _parse_keyword(self, keyword):
        if self._parse_token() == keyword:
            return self._token
        else:
            self._parse_unexpected("semicolon" if keyword == ";" else keyword)

    def _parse_keywords(self, keywords):
        if self._parse_token() in keywords:
            return self._token
        else:
            self._parse_unexpected("one of {}".format(", ".join(keywords)))

    def _parse_value(self, kind):
        if isinstance(self._parse_token(), kind):
            return self._token
        else:
            if kind == int:
                expected = "integer"
            elif kind == float:
                expected = "real"
            elif kind == (int, float):
                expected = "number"
            elif kind == bits:
                expected = "scan data"
            elif kind == tuple:
                expected = "data"
            else:
                assert False
            self._parse_unexpected(expected)

    def _parse_trst_mode(self):
        if self._parse_token() in _trst_modes:
            return self._token
        else:
            self._parse_unexpected("TRST mode", _trst_modes)

    def _parse_tap_state(self):
        if self._parse_token() in _tap_states:
            return self._token
        else:
            self._parse_unexpected("TAP state", _tap_states)

    def _parse_tap_stable_state(self):
        if self._parse_token() in _tap_stable_states:
            return self._token
        else:
            self._parse_unexpected("stable TAP state", _tap_stable_states)

    def _parse_scan_data(self, length):
        value = self._parse_value(bits)
        if int(value[length:]) != 0:
            self._parse_error("scan data length %d exceeds command length %d"
                              % (len(value), length))

        if length > len(value):
            return value + bits(0, length - len(value))
        else:
            return value[:length]

    def parse_command(self):
        self._cmd_pos = self._lexer.position

        command = self._parse_token()
        if command is None:
            return False

        elif command == "FREQUENCY":
            cycles = self._try(self._parse_value, (int, float))
            if cycles is not None:
                self._parse_keyword("HZ")
            self._parse_keyword(";")

            result = self._handler.svf_frequency(frequency=cycles)

        elif command == "TRST":
            mode = self._parse_trst_mode()
            self._parse_keyword(";")

            result = self._handler.svf_trst(mode=mode)

        elif command == "STATE":
            states = []
            while True:
                state = self._try(self._parse_tap_state)
                if state is None: break
                states.append(state)

            self._parse_keyword(";")

            if not states:
                self._parse_error("at least one state required")
            if states[-1] not in _tap_stable_states:
                self._parse_error("last state must be a stable state")

            path_states, stable_state = states[0:-1], states[-1]
            result = self._handler.svf_state(state=stable_state, path=path_states)

        elif command in ("ENDIR", "ENDDR"):
            stable_state = self._parse_tap_stable_state()
            self._parse_keyword(";")

            if command == "ENDIR":
                result = self._handler.svf_endir(state=stable_state)
            if command == "ENDDR":
                result = self._handler.svf_enddr(state=stable_state)

        elif command in ("HIR", "SIR", "TIR", "HDR", "SDR", "TDR"):
            length = self._parse_value(int)

            if self._param_mask[command] is None or len(self._param_mask[command]) != length:
                self._param_mask[command] = bits(-1, length)
            if self._param_smask[command] is None or len(self._param_smask[command]) != length:
                self._param_smask[command] = bits(-1, length)

            param_tdi   = self._param_tdi[command]
            param_tdo   = None
            param_mask  = self._param_mask[command]
            param_smask = self._param_smask[command]
            parameters  = set()
            while True:
                parameter = self._try(self._parse_keywords, ("TDI", "TDO", "MASK", "SMASK"))
                if parameter is None: break

                value = self._parse_scan_data(length)
                if parameter in parameters:
                    self._parse_error("parameter %s specified twice" % parameter)
                parameters.add(parameter)

                if parameter == "TDI":
                    self._param_tdi[command] = value
                    param_tdi = value
                if parameter == "TDO":
                    param_tdo = value
                if parameter == "MASK":
                    self._param_mask[command] = value
                    param_mask = value
                if parameter == "SMASK":
                    self._param_smask[command] = value
                    param_smask = value

            self._parse_keyword(";")

            if param_tdi is None and length == 0:
                param_tdi = bits()
            elif param_tdi is None:
                self._parse_error("initial value for parameter TDI required")
            if len(param_tdi) != length:
                self._parse_error("parameter TDI needs to be specified again because "
                                  "the length changed")

            if param_tdo is None:
                # Make it a bit easier for downstream; set MASK (but not remembered MASK)
                # to "all don't care" if there's no TDO specified.
                param_mask = bits(0, len(param_mask))

            if command == "HIR":
                result = self._handler.svf_hir(tdi=param_tdi, smask=param_smask,
                                               tdo=param_tdo,  mask=param_mask)
            if command == "SIR":
                result = self._handler.svf_sir(tdi=param_tdi, smask=param_smask,
                                               tdo=param_tdo,  mask=param_mask)
            if command == "TIR":
                result = self._handler.svf_tir(tdi=param_tdi, smask=param_smask,
                                               tdo=param_tdo,  mask=param_mask)
            if command == "HDR":
                result = self._handler.svf_hdr(tdi=param_tdi, smask=param_smask,
                                               tdo=param_tdo,  mask=param_mask)
            if command == "SDR":
                result = self._handler.svf_sdr(tdi=param_tdi, smask=param_smask,
                                               tdo=param_tdo,  mask=param_mask)
            if command == "TDR":
                result = self._handler.svf_tdr(tdi=param_tdi, smask=param_smask,
                                               tdo=param_tdo,  mask=param_mask)

        elif command == "RUNTEST":
            run_state = self._try(self._parse_tap_stable_state)
            run_params = self._try(lambda:
                (self._parse_value(int), self._parse_keywords(("TCK", "SCK"))))
            if run_params is None:
                run_count, run_clock = None, "TCK"
                min_time, _ = \
                     self._parse_value((int, float)), self._parse_keyword("SEC")
            else:
                run_count, run_clock = run_params
                min_time, _ = self._try(lambda:
                    (self._parse_value((int, float)), self._parse_keyword("SEC"))) \
                    or (None, None)
            if self._try(self._parse_keyword, "MAXIMUM"):
                max_time, _ = \
                     self._parse_value((int, float)), self._parse_keyword("SEC")
            else:
                max_time = None
            if self._try(self._parse_keyword, "ENDSTATE"):
                end_state = self._parse_tap_stable_state()
            else:
                end_state = None
            self._parse_keyword(";")

            if run_state is None:
                run_state = self._param_run_state
            else:
                self._param_run_state = run_state
                if end_state is None:
                    end_state = run_state

            if end_state is None:
                end_state = self._param_end_state
            else:
                self._param_end_state = end_state

            if run_clock is None:
                run_clock = "TCK"

            if max_time is not None and min_time is not None and max_time < min_time:
                self._parse_error("maximum time must be greater than minimum time")

            result = self._handler.svf_runtest(run_state=run_state,
                                      run_count=run_count, run_clock=run_clock,
                                      min_time =min_time,   max_time=max_time,
                                      end_state=end_state)

        elif command == "PIOMAP":
            mapping, = self._parse_value(tuple)
            self._parse_keyword(";")

            result = self._handler.svf_piomap(mapping=mapping)

        elif command == "PIO":
            vector, = self._parse_value(tuple)
            self._parse_keyword(";")

            result = self._handler.svf_pio(vector=vector)

        else:
            self._parse_unexpected("command", _commands)

        return result or True

    def last_command(self):
        return self._lexer.buffer[self._cmd_pos:self._lexer.position]

    def parse_file(self):
        while self.parse_command(): pass


class SVFEventHandler(metaclass=ABCMeta):
    """
    An abstract base class for Serial Vector Format parsing events.

    The methods of this class are called when a well-formed SVF command is encountered.
    The parser takes care of maintaining all lexical state (e.g. "sticky" parameters),
    but all logical state is maintained by the event handler.
    """

    @abstractmethod
    def svf_frequency(self, frequency):
        """Called when the ``FREQUENCY`` command is encountered."""

    @abstractmethod
    def svf_trst(self, mode):
        """Called when the ``TRST`` command is encountered."""

    @abstractmethod
    def svf_state(self, state, path):
        """Called when the ``STATE`` command is encountered."""

    @abstractmethod
    def svf_endir(self, state):
        """Called when the ``ENDIR`` command is encountered."""

    @abstractmethod
    def svf_enddr(self, state):
        """Called when the ``ENDDR`` command is encountered."""

    @abstractmethod
    def svf_hir(self, tdi, smask, tdo, mask):
        """Called when the ``HIR`` command is encountered."""

    @abstractmethod
    def svf_sir(self, tdi, smask, tdo, mask):
        """Called when the ``SIR`` command is encountered."""

    @abstractmethod
    def svf_tir(self, tdi, smask, tdo, mask):
        """Called when the ``TIR`` command is encountered."""

    @abstractmethod
    def svf_hdr(self, tdi, smask, tdo, mask):
        """Called when the ``HDR`` command is encountered."""

    @abstractmethod
    def svf_sdr(self, tdi, smask, tdo, mask):
        """Called when the ``SDR`` command is encountered."""

    @abstractmethod
    def svf_tdr(self, tdi, smask, tdo, mask):
        """Called when the ``TDR`` command is encountered."""

    @abstractmethod
    def svf_runtest(self, run_state, run_count, run_clock, min_time, max_time, end_state):
        """Called when the ``RUNTEST`` command is encountered."""

    @abstractmethod
    def svf_piomap(self, mapping):
        """Called when the ``PIOMAP`` command is encountered."""

    @abstractmethod
    def svf_pio(self, vector):
        """Called when the ``PIO`` command is encountered."""

# -------------------------------------------------------------------------------------------------

import unittest


class SVFLexerTestCase(unittest.TestCase):
    def assertLexes(self, source, tokens):
        self.lexer = SVFLexer(source)
        self.assertEqual(list(self.lexer), tokens)

    def test_eof(self):
        self.assertLexes("", [])

    def test_comment(self):
        self.assertLexes("!foo",
                         [])
        self.assertLexes("//foo",
                         [])
        self.assertLexes("//foo\n!bar\n",
                         [])
        self.assertLexes("//foo\n!bar\nTRST",
                         ["TRST"])

    def test_keyword(self):
        self.assertLexes("TRST",
                         ["TRST"])
        self.assertLexes("TRST OFF;",
                         ["TRST", "OFF", ";"])

    def test_integer(self):
        self.assertLexes("8",       [8])
        self.assertLexes("12",      [12])

    def test_real(self):
        self.assertLexes("1E6",     [1e6])
        self.assertLexes("1E+6",    [1e6])
        self.assertLexes("1E-6",    [1e-6])
        self.assertLexes("1.1E6",   [1.1e6])
        self.assertLexes("1.1",     [1.1])

    def test_bits(self):
        self.assertLexes("(0)",        [bits("")])
        self.assertLexes("(1)",        [bits("1")])
        self.assertLexes("(F)",        [bits("1111")])
        self.assertLexes("(f)",        [bits("1111")])
        self.assertLexes("(0F)",       [bits("1111")])
        self.assertLexes("(A\n5)",     [bits("10100101")]) # Test literals split over two lines
        self.assertLexes("(A\n\t5)",   [bits("10100101")]) # With potential whitespace
        self.assertLexes("(A\n  5)",   [bits("10100101")])
        self.assertLexes("(A\r\n5)",   [bits("10100101")]) # Support both LF & LFCR
        self.assertLexes("(A\r\n\t5)", [bits("10100101")])
        self.assertLexes("(A\r\n  5)", [bits("10100101")])
        self.assertLexes("(FF)",       [bits("11111111")])
        self.assertLexes("(1AA)",      [bits("110101010")])

    def test_literal(self):
        self.assertLexes("(HHZZL)",     [("HHZZL",)])
        self.assertLexes("(IN FOO)",    [("IN FOO",)])

    def test_error(self):
        with self.assertRaises(SVFParsingError):
            SVFLexer("XXX").next()


class SVFMockEventHandler:
    def __init__(self):
        self.events = []

    def __getattr__(self, name):
        if name.startswith("svf_"):
            def svf_event(**kwargs):
                self.events.append((name, kwargs))
            return svf_event
        else:
            return super().__getattr__(name)


class SVFParserTestCase(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def assertParses(self, source, events):
        self.handler = SVFMockEventHandler()
        self.parser = SVFParser(source, self.handler)
        self.parser.parse_file()
        self.assertEqual(self.handler.events, events)

    def assertErrors(self, source, error):
        with self.assertRaisesRegex(SVFParsingError, r"^{}".format(re.escape(error))):
            self.handler = SVFMockEventHandler()
            self.parser = SVFParser(source, self.handler)
            self.parser.parse_file()

    def test_frequency(self):
        self.assertParses("FREQUENCY;",
                          [("svf_frequency", {"frequency": None})])
        self.assertParses("FREQUENCY 1E6 HZ;",
                          [("svf_frequency", {"frequency": 1e6})])
        self.assertParses("FREQUENCY 1000 HZ;",
                          [("svf_frequency", {"frequency": 1000})])

        self.assertErrors("FREQUENCY 1E6;",
                          "expected HZ")

    def test_trst(self):
        self.assertParses("TRST ON;",
                          [("svf_trst", {"mode": "ON"})])
        self.assertParses("TRST OFF;",
                          [("svf_trst", {"mode": "OFF"})])
        self.assertParses("TRST Z;",
                          [("svf_trst", {"mode": "Z"})])
        self.assertParses("TRST ABSENT;",
                          [("svf_trst", {"mode": "ABSENT"})])

        self.assertErrors("TRST HZ;",
                          "expected TRST mode")

    def test_state(self):
        self.assertParses("STATE IDLE;",
                          [("svf_state", {"state": "IDLE", "path": []})])
        self.assertParses("STATE IRUPDATE IDLE;",
                          [("svf_state", {"state": "IDLE", "path": ["IRUPDATE"]})])
        self.assertParses("STATE IREXIT2 IRUPDATE IDLE;",
                          [("svf_state", {"state": "IDLE", "path": ["IREXIT2", "IRUPDATE"]})])

        self.assertErrors("STATE;",
                          "at least one state required")
        self.assertErrors("STATE IRSHIFT;",
                          "last state must be a stable state")
        self.assertErrors("STATE RESET IRSHIFT;",
                          "last state must be a stable state")

    def test_endir_enddr(self):
        for command, event in [
            ("ENDIR", "svf_endir"),
            ("ENDDR", "svf_enddr")
        ]:
            self.assertParses("{c} IRPAUSE;".format(c=command),
                              [(event, {"state": "IRPAUSE"})])

            self.assertErrors("{c} IRSHIFT;".format(c=command),
                              "expected stable TAP state")
            self.assertErrors("{c};".format(c=command),
                              "expected stable TAP state")

    def test_hir_sir_tir_hdr_sdr_tdr(self):
        for command, event in [
            ("HIR", "svf_hir"),
            ("SIR", "svf_sir"),
            ("TIR", "svf_tir"),
            ("HDR", "svf_hdr"),
            ("SDR", "svf_sdr"),
            ("TDR", "svf_tdr"),
        ]:
            self.assertParses("{c} 0;".format(c=command), [
                (event, {
                    "tdi":   bits(""),
                    "smask": bits(""),
                    "tdo":   None,
                    "mask":  bits(""),
                }),
            ])
            self.assertParses("{c} 8 TDI(a);".format(c=command), [
                (event, {
                    "tdi":   bits("00001010"),
                    "smask": bits("11111111"),
                    "tdo":   None,
                    "mask":  bits("00000000"),
                }),
            ])
            self.assertParses("{c} 6 TDI(0a);".format(c=command), [
                (event, {
                    "tdi":   bits("001010"),
                    "smask": bits("111111"),
                    "tdo":   None,
                    "mask":  bits("000000"),
                }),
            ])
            self.assertParses("{c} 8 TDI(a); {c} 8;".format(c=command), [
                (event, {
                    "tdi":   bits("00001010"),
                    "smask": bits("11111111"),
                    "tdo":   None,
                    "mask":  bits("00000000"),
                }),
                (event, {
                    "tdi":   bits("00001010"),
                    "smask": bits("11111111"),
                    "tdo":   None,
                    "mask":  bits("00000000"),
                }),
            ])
            self.assertParses("{c} 8 TDI(a) SMASK(3); {c} 8; {c} 12 TDI(b);"
                              .format(c=command), [
                (event, {
                    "tdi":   bits("00001010"),
                    "smask": bits("00000011"),
                    "tdo":   None,
                    "mask":  bits("00000000"),
                }),
                (event, {
                    "tdi":   bits("00001010"),
                    "smask": bits("00000011"),
                    "tdo":   None,
                    "mask":  bits("00000000"),
                }),
                (event, {
                    "tdi":   bits("000000001011"),
                    "smask": bits("111111111111"),
                    "tdo":   None,
                    "mask":  bits("000000000000"),
                }),
            ])
            self.assertParses("{c} 8 TDI(0) TDO(a) MASK(3); {c} 8; "
                              "{c} 8 TDO(1); {c} 12 TDI(0) TDO(b);"
                              .format(c=command), [
                (event, {
                    "tdi":   bits("00000000"),
                    "smask": bits("11111111"),
                    "tdo":   bits("00001010"),
                    "mask":  bits("00000011"),
                }),
                (event, {
                    "tdi":   bits("00000000"),
                    "smask": bits("11111111"),
                    "tdo":   None,
                    "mask":  bits("00000000"),
                }),
                (event, {
                    "tdi":   bits("00000000"),
                    "smask": bits("11111111"),
                    "tdo":   bits("00000001"),
                    "mask":  bits("00000011"),
                }),
                (event, {
                    "tdi":   bits("000000000000"),
                    "smask": bits("111111111111"),
                    "tdo":   bits("000000001011"),
                    "mask":  bits("111111111111"),
                }),
            ])

            self.assertErrors("{c} 8 TDI(aaa);".format(c=command),
                              "scan data length 12 exceeds command length 8")
            self.assertErrors("{c} 8 TDI(0) TDI(0);".format(c=command),
                              "parameter TDI specified twice")
            self.assertErrors("{c} 8;".format(c=command),
                              "initial value for parameter TDI required")
            self.assertErrors("{c} 8 TDI(aa); {c} 12;".format(c=command),
                              "parameter TDI needs to be specified again because "
                              "the length changed")

    def test_runtest(self):
        self.assertParses("RUNTEST 20000 TCK;", [
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": 20000, "run_clock": "TCK",
                "min_time":  None,    "max_time":  None,  "end_state": "IDLE"
            }),
        ])
        self.assertParses("RUNTEST 20000 TCK 1E3 SEC;", [
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": 20000, "run_clock": "TCK",
                "min_time":  1e3,     "max_time":  None,  "end_state": "IDLE"
            }),
        ])
        self.assertParses("RUNTEST 20000 TCK 1E3 SEC MAXIMUM 1E6 SEC;", [
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": 20000, "run_clock": "TCK",
                "min_time":  1e3,     "max_time":  1e6,   "end_state": "IDLE"
            }),
        ])
        self.assertParses("RUNTEST 20000 TCK 1E3 SEC MAXIMUM 1E6 SEC ENDSTATE RESET;", [
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": 20000, "run_clock": "TCK",
                "min_time":  1e3,     "max_time":  1e6,   "end_state": "RESET"
            }),
        ])
        self.assertParses("RUNTEST 20000 TCK 1E3 SEC MAXIMUM 1E6 SEC ENDSTATE RESET;", [
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": 20000, "run_clock": "TCK",
                "min_time":  1e3,     "max_time":  1e6,   "end_state": "RESET"
            }),
        ])
        self.assertParses("RUNTEST 20000 TCK ENDSTATE RESET; RUNTEST 100 TCK;", [
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": 20000, "run_clock": "TCK",
                "min_time":  None,    "max_time":  None,  "end_state": "RESET"
            }),
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": 100,   "run_clock": "TCK",
                "min_time":  None,    "max_time":  None,  "end_state": "RESET"
            }),
        ])
        self.assertParses("RUNTEST RESET 20000 TCK ENDSTATE RESET; RUNTEST IDLE 100 TCK;", [
            ("svf_runtest", {
                "run_state": "RESET", "run_count": 20000, "run_clock": "TCK",
                "min_time":  None,    "max_time":  None,  "end_state": "RESET"
            }),
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": 100,   "run_clock": "TCK",
                "min_time":  None,    "max_time":  None,  "end_state": "IDLE"
            }),
        ])

        self.assertParses("RUNTEST 20000 SCK;", [
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": 20000, "run_clock": "SCK",
                "min_time":  None,    "max_time":  None,  "end_state": "IDLE"
            }),
        ])

        self.assertParses("RUNTEST 1 SEC;", [
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": None,  "run_clock": "TCK",
                "min_time":  1,       "max_time":  None,  "end_state": "IDLE"
            }),
        ])
        self.assertParses("RUNTEST 1 SEC MAXIMUM 2 SEC;", [
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": None,  "run_clock": "TCK",
                "min_time":  1,       "max_time":  2,     "end_state": "IDLE"
            }),
        ])
        self.assertParses("RUNTEST 200 TCK ENDSTATE RESET; RUNTEST 1 SEC;", [
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": 200,   "run_clock": "TCK",
                "min_time":  None,    "max_time":  None,  "end_state": "RESET"
            }),
            ("svf_runtest", {
                "run_state": "IDLE",  "run_count": None,  "run_clock": "TCK",
                "min_time":  1,       "max_time":  None,  "end_state": "RESET"
            }),
        ])

        self.assertErrors("RUNTEST;",
                          "expected number")
        self.assertErrors("RUNTEST 2 SEC MAXIMUM 1 SEC;",
                          "maximum time must be greater than minimum time")

    def test_piomap(self):
        self.assertParses("PIOMAP (IN FOO OUT BAR);",
                          [("svf_piomap", {"mapping": "IN FOO OUT BAR"})])

        self.assertErrors("PIOMAP;",
                          "expected data")

    def test_pio(self):
        self.assertParses("PIO (LHZX);",
                          [("svf_pio", {"vector": "LHZX"})])

        self.assertErrors("PIO;",
                          "expected data")

    def test_last_command(self):
        handler = SVFMockEventHandler()
        parser = SVFParser(" TRST OFF; SIR 8 TDI (aa); ", handler)
        parser.parse_command()
        self.assertEqual(parser.last_command(), " TRST OFF;")
        parser.parse_command()
        self.assertEqual(parser.last_command(), " SIR 8 TDI (aa);")

# -------------------------------------------------------------------------------------------------

class SVFPrintingEventHandler:
    def __getattr__(self, name):
        if name.startswith("svf_"):
            def svf_event(**kwargs):
                print((name, kwargs))
            return svf_event
        else:
            return super().__getattr__(name)


if __name__ == "__main__":
    import sys
    with open(sys.argv[1]) as f:
        SVFParser(f.read(), SVFPrintingEventHandler()).parse_file()



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

# Vendor requests.
REQUEST_DEBUG_SPI_SEND          = 0x50
REQUEST_DEBUG_SPI_READ_RESPONSE = 0x51


class DebugSPIConnection:
    """ Connection to the FPGA via Apollo's Debug SPI. """

    def __init__(self, debugger):
        self._debugger = debugger

        self.command_bytes  = None
        self.register_bytes = None
        self.chunk_size     = 256 + 4


    def _transfer_chunk(self, data_to_send, *, complete=True, invert_cs=False):
        """ Transfers a set of data over SPI, and reads the response. """

        # Transfer the data to be sent...
        self._debugger.out_request(REQUEST_DEBUG_SPI_SEND,
            data=data_to_send,
            value=0 if complete else 1,
            index=1 if invert_cs else 0
        )

        # ... and read the response.
        return self._debugger.in_request(REQUEST_DEBUG_SPI_READ_RESPONSE, length=len(data_to_send))


    def transfer(self, data_to_send, invert_cs=False):
        """ Transfers a set of data over SPI, and reads the response.

        Parameters:
            data_to_send -- The data to be sent; also sets the length of received data.
            invert_cs    -- Perform the transaction with CS high, rather than low.
                            Useful for multiplexing two targets on the same SPI bus
                            with a single CS line.
        """

        to_send  = bytearray(data_to_send)
        response = bytearray()

        while to_send:
            chunk = to_send[0:self.chunk_size]
            del to_send[0:self.chunk_size]

            response_chunk = \
                self._transfer_chunk(chunk, complete=not to_send, invert_cs=invert_cs)
            response.extend(response_chunk)

        return response


    def _autodetect_command_shape(self):

        # Send out a chain of 16 zeroes, and see what we get back.
        autodetect_value = int.from_bytes(self.transfer(b"\x00" * 16), byteorder="big")

        # Get the start of the transaction; e.g. the sequence of zeroes followed by the sequence of ones.
        autodetect_bits  = "{:0128b}".format(autodetect_value).rstrip('0')

        # The leftover number of 0's is our command size in bits; the address size is one less than that.
        # The leftover number of 1's is our register size in bits.
        command_bits        = autodetect_bits.count('0')
        register_bits       = autodetect_bits.count('1')
        self.command_bytes  = command_bits // 8
        self.register_bytes = register_bits // 8

        # Sanity check our command-shape detection.
        invalid_shape = \
            (self.command_bytes  == 0) or \
            (self.register_bytes == 0) or \
            (command_bits  % 8   != 0) or \
            (register_bits % 8   != 0)

        if invalid_shape:
            raise IOError("Failed to autonegotiate SPI address/register size.")


    def register_transaction(self, address, *, is_write, value=0):
        """ Performs an SPI register transaction. """

        if (self.command_bytes is None) or (self.register_bytes is None):
            self._autodetect_command_shape()

        # Compute our write flag.
        write_flag_position = (self.command_bytes * 8) - 1
        write_flag          = (1 << write_flag_position) if is_write else 0

        # Compute our command and value words.
        command = (write_flag | address).to_bytes(self.command_bytes, byteorder='big')
        value   = value.to_bytes(self.register_bytes, byteorder='big')

        # Finally, issue our transaction...
        raw_response = self.transfer(command + value)

        # ... and convert the response back into an integer.
        raw_value = raw_response[self.command_bytes:]
        return int.from_bytes(raw_value, byteorder='big')


    def register_read(self, address):
        """ Reads a value from the provided registers."""
        return self.register_transaction(address, is_write=False)

    def register_write(self, address, value):
        """ Writes a value to the provided register. """
        return self.register_transaction(address, value=value, is_write=True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['apollo_fpga',
 'apollo_fpga.commands',
 'apollo_fpga.protocol',
 'apollo_fpga.support']

package_data = \
{'': ['*']}

install_requires = \
['pyusb>=1.1.1,<2.0.0', 'pyvcd>=0.2.4,<0.3.0']

entry_points = \
{'console_scripts': ['apollo = apollo_fpga.commands.cli:main']}

setup_kwargs = {
    'name': 'apollo-fpga',
    'version': '0.0.1',
    'description': 'host tools for Apollo FPGA debug controllers',
    'long_description': None,
    'author': 'Katherine Temkin',
    'author_email': 'k@ktemkin.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Helpers for building self-test applets around Apollo."""

import sys
import inspect
import logging

from prompt_toolkit import HTML
from prompt_toolkit import print_formatted_text as pprint

from .. import ApolloDebugger


def named_test(name):
    """ Decorator that applies a name to an Apollo test case. """

    # Create a simple decorator that applies the relevant name.
    def decorator(func):
        func.__name__ = name
        return func

    return decorator


class ApolloSelfTestCase:
    """ Self-test running for simple self-tests via Apollo.

    Automatically creates a debugger, and then executes any function with
    a name that starts with `test_`, passing in the debugger connection.
    """

    def fail(self, explanation: str ="called fail()"):
        """ Called to indicate that a test case failed. """
        raise AssertionError(explanation)


    def assertRegisterValue(self, number: int, expected_value: int):
        """ Asserts that a DUT register has a given value. """

        actual_value = self.dut.registers.register_read(number)

        if actual_value != expected_value:
            raise AssertionError(f"register {number} was {actual_value}, not expected {expected_value}")


    def _run_as_test_case(self, method: callable, dut: ApolloDebugger):
        """ Runs a given method as a test case. """

        # Try to run the method...
        try:
            method(dut)

            # If nothing went wrong, succeed.
            result = "<green>✓ PASSED</green>"
            self.successes += 1
        except AssertionError as e:
            result = f"<red>✗ FAILED</red>\t({e})"
            self.failures += 1
        except Exception as e:
            result = f"<red>⚠ EXCEPTION </red>\t({e})"
            self.exceptions += 1

        # ... and print the result.
        name = method.__name__.replace("test_", "")
        pprint(HTML(f"\t{name}:\t{result}"))


    def run_tests(self):
        """ Runs all tests in the given test-case."""

        self.successes  = 0
        self.failures   = 0
        self.exceptions = 0

        # Create our device under test connection.
        dut = self.dut = ApolloDebugger()
        logging.info(f"Connected to onboard debugger; hardware revision r{dut.major}.{dut.minor} (s/n: {dut.serial_number}).")

        # Find all test methods attached to this object...
        logging.info("Running tests...")

        sys.stdout.flush()
        sys.stderr.flush()

        with dut.jtag as jtag:

            pprint(HTML("\n\n<b><u>Automated tests:</u></b>"))
            for name, member in inspect.getmembers(self):
                if inspect.ismethod(member) and name.startswith('test_'):
                    self._run_as_test_case(member, self.dut)




#
# This file is part of LUNA
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" ECP5 configuration code for LUNA. """

import time

from enum import IntEnum
from collections import defaultdict

from .jtag import JTAGChain
from .support.bits import bits


class IntelJTAGProgrammer:
    """ Apollo JTAG configuration tool for Intel FPGAs.

    Parameters
    ----------
    jtag_chain: JTAGChain
        The JTAG chain to be used for configuration.
    """

    def __init__(self, jtag_chain, *args, **kwargs):

        # Store a reference to our board and SPI bus.
        self.chain = jtag_chain


    def _generate_bit_reversed_bitstream(self, bitstream):
        """
        Generates a copy of the provided bitstream with the bits in each byte
        reversed -- in the format the FPGA likes them for MSPI mode.
        """

        # Quick helper function to reverse the bits in our bitstream.
        def reverse_bits(num):
            binstr = "{:08b}".format(num)
            return int(binstr[::-1], 2)

        # Reverse each of the bits in each byte of the bitstream.
        #
        # This ensures that bits are shifted into the FPGA in the same
        # order as they need to be presented to the configuration logic;
        # even if the FPGA is the one commanding the flash.
        #
        bit_reversed = bytearray(bitstream)
        for i in range(len(bit_reversed)):
            bit_reversed[i] = reverse_bits(bit_reversed[i])

        return bit_reversed



    def configure(self, bitstream):
        """ Configures the FPGA. """

        bitstream_length_bits = len(bitstream) * 8

        # The actual JTAG commands for loading bitstreams are not clearly documented;
        # these values were found by copying the least collection of instructions that
        # seemed to result in a reliable configuration from a Quartus-generated SVF,
        # and correlating with CYIV-51008, which references some instruction ordering
        # and names without additional documentation.
        #
        # These could be very wrong.
        #
        IR_LENGTH         = 10
        JTAG_PROGRAM      = 0x002
        JTAG_STARTUP      = 0x003

        # Upload our bitstream to our board ...
        self.chain.shift_instruction(JTAG_PROGRAM, length=IR_LENGTH, state_after='IRPAUSE')
        self.chain.shift_data(tdi=bitstream, length=bitstream_length_bits,
            ignore_response=True, state_after='IDLE', byteorder='little')

        # ... and apply it / let it start.
        self.chain.shift_instruction(JTAG_STARTUP, length=IR_LENGTH, state_after='IRPAUSE')
        self.chain.run_test(102400)

